from typing import Optional, List, Union, Mapping, Tuple

import torch.jit
from torch import nn, Tensor

from pytorch_toolbelt.modules.interfaces import AbstractHead, FeatureMapsSpecification

__all__ = ["ResizeHead"]


class ResizeHead(AbstractHead):
    """
    Basic prediction head that applies a dropout and convolution (without normalization and activation) to the
    feature map with the smallest stride and resize the predicted feature map to the input image size.
    """

    def __init__(
        self,
        input_spec: FeatureMapsSpecification,
        num_classes: int,
        output_name: Optional[str] = None,
        kernel_size: int = 3,
        dropout_rate: float = 0.0,
        dropout_inplace: bool = False,
        interpolation_mode="bilinear",
        interpolation_align_corners=False,
    ):
        """
        Initialize prediction head

        :param input_spec: Specification of input feature maps
        :param num_classes: Number of classes to predict
        :param output_name: Name of the output tensor. If None, returns tensor directly, otherwise returns dict with { output_name: tensor }
        :param kernel_size: Convolution kernel size. Padding is automatically computed to using kernel_size // 2 formula.
        :param dropout_rate: Dropout rate to apply before convolution.
        :param dropout_inplace: If True, applies dropout in-place.
        :param interpolation_mode: Interpolation mode to use for upsampling. See torch.nn.functional.interpolate for details.
        :param interpolation_align_corners: Interpolation align corners mode to use for upsampling. See torch.nn.functional.interpolate for details.
        """

        super().__init__(input_spec)
        self.target_feature_map_index = input_spec.get_index_of_largest_feature_map()
        self.output_name = output_name

        channels = input_spec.channels[self.target_feature_map_index]

        self.dropout = nn.Dropout2d(dropout_rate, inplace=dropout_inplace)
        self.final = nn.Conv2d(channels, num_classes, kernel_size=kernel_size, padding=kernel_size // 2, bias=True)

        self.output_spec = FeatureMapsSpecification(channels=(num_classes,), strides=(1,))

        self.interpolation_mode = interpolation_mode
        self.interpolation_align_corners = interpolation_align_corners

    def forward(
        self, feature_maps: List[Tensor], output_size: Union[Tuple[int, int], torch.Size]
    ) -> Union[Tensor, Tuple[Tensor, ...], List[Tensor], Mapping[str, Tensor]]:
        """

        :param feature_maps: Input features (from encoder, decoder or neck, etc.)
        :param output_size:  The desired output size (rows, cols) of the output feature map.
        :return:
        """
        x = feature_maps[self.target_feature_map_index]
        x = self.dropout(x)
        x = self.final(x)

        output = torch.nn.functional.interpolate(
            x, size=output_size, mode=self.interpolation_mode, align_corners=self.interpolation_align_corners
        )

        if self.output_name is not None:
            return {self.output_name: output}
        else:
            return output

    @torch.jit.unused
    def get_output_spec(self) -> FeatureMapsSpecification:
        return self.output_spec



from typing import List, Dict, Optional, Callable

import torch
import torch.nn as nn

from pytorch_toolbelt.datasets import name_for_stride
from pytorch_toolbelt.modules.simple import conv1x1

from pytorch_toolbelt.modules.interfaces import AbstractHead, FeatureMapsSpecification

__all__ = ["DeepSupervisionHead"]


class DeepSupervisionHead(AbstractHead):
    """
    Deep supervision head takes all feature maps and applies 1x1 convolution to each of them.
    """

    __constants__ = ["output_name_prefix", "output_spec"]

    def __init__(
        self, input_spec: FeatureMapsSpecification, num_classes: int, output_name_prefix: Optional[str] = None
    ):
        super().__init__(input_spec)
        self.heads = nn.ModuleList([conv1x1(channel, num_classes) for channel in input_spec.channels])
        self.output_spec = FeatureMapsSpecification(
            channels=tuple(
                [num_classes] * len(input_spec.channels),
            ),
            strides=input_spec.strides,
        )
        self.output_name_prefix = output_name_prefix

    def forward(self, feature_maps: List[torch.Tensor], output_size=None) -> Dict[str, torch.Tensor]:
        if self.output_name_prefix is None:
            outputs = []
            for feature_map, head, output_stride in zip(feature_maps, self.heads, self.output_spec.strides):
                output = head(feature_map)
                outputs.append(output)
        else:
            outputs = {}
            for feature_map, head, output_stride in zip(feature_maps, self.heads, self.output_spec.strides):
                output = head(feature_map)
                outputs[name_for_stride(self.output_name_prefix, output_stride)] = output

        return outputs

    @torch.jit.unused
    def get_output_spec(self) -> FeatureMapsSpecification:
        return self.output_spec

    def apply_to_final_layer(self, func: Callable[[nn.Module], None]):
        for head in self.heads:
            func(head)



import inspect
from typing import List, Union, Callable, Type

import torch
from torch import Tensor, nn

from pytorch_toolbelt.modules.interfaces import AbstractDecoder, FeatureMapsSpecification
from .. import conv1x1, conv3x3, instantiate_upsample_block, UpsampleLayerType, AbstractResizeLayer

__all__ = ["FPNDecoder"]


class FPNDecoder(AbstractDecoder):
    """
    Feature pyramid network decoder with summation between intermediate layers:

        Input
        feature_map[0] -> bottleneck[0](feature_map[0]) + upsample(fpn[1]) -> fpn[0]
        feature_map[1] -> bottleneck[1](feature_map[1]) + upsample(fpn[2]) -> fpn[1]
        feature_map[2] -> bottleneck[2](feature_map[2]) + upsample(fpn[3]) -> fpn[2]
        ...
        feature_map[n-1] -> bottleneck[n-1](feature_map[n-1]) + upsample(context) -> fpn[n]
        feature_map[n] -> context_block(feature_map[n]) -> context
    """

    def __init__(
        self,
        input_spec: FeatureMapsSpecification,
        out_channels: int,
        bottleneck_block: Callable[[int, int], nn.Module] = conv1x1,
        prediction_block: Union[nn.Identity, Callable[[int, int], nn.Module]] = conv3x3,
        upsample_block: Union[UpsampleLayerType, Type[AbstractResizeLayer]] = UpsampleLayerType.BILINEAR,
    ):
        """
        Create a new instance of FPN decoder with summation of consecutive feature maps.
        :param out_channels: Number of output channels in each feature map
        :param bottleneck_block:
        :param prediction_block: Optional prediction block to apply to FPN feature maps before returning from decoder
        :param prediction_channels: Number of prediction channels
        :param upsample_block:
        """
        super().__init__(input_spec)

        feature_maps = input_spec.channels

        self.lateral = nn.ModuleList([bottleneck_block(in_channels, out_channels) for in_channels in feature_maps])

        if inspect.isclass(prediction_block) and issubclass(prediction_block, nn.Identity):
            self.outputs = nn.ModuleList([prediction_block() for _ in feature_maps[:-1]])
        else:
            self.outputs = nn.ModuleList([prediction_block(out_channels, out_channels) for _ in feature_maps[:-1]])

        # TODO: Check that upsample_block we using produces the same number of output channels as input
        # TODO: Calculate the scale_factor form input_spec.strides
        num_upsample_blocks = len(feature_maps) - 1
        self.upsamples = nn.ModuleList(
            [
                instantiate_upsample_block(upsample_block, in_channels=out_channels, scale_factor=2)
                for _ in range(num_upsample_blocks)
            ]
        )

        self.output_spec = FeatureMapsSpecification(
            channels=[out_channels] * len(feature_maps), strides=input_spec.strides
        )

    @torch.jit.unused
    def get_output_spec(self) -> FeatureMapsSpecification:
        return self.output_spec

    def forward(self, feature_maps: List[Tensor]) -> List[Tensor]:
        # Lateral connections
        lateral_maps = [lateral(feature_map) for feature_map, lateral in zip(feature_maps, self.lateral)]

        last_feature_map = lateral_maps[-1]
        remaining_feature_maps = lateral_maps[:-1][::-1]

        outputs = [last_feature_map]

        for feature_map, upsample, output_block in zip(remaining_feature_maps, self.upsamples, self.outputs):
            upsampled = upsample(outputs[-1], output_size=feature_map.shape[-2:])

            fpn = output_block(feature_map + upsampled)
            outputs.append(fpn)

        # Returns list of tensors in same order as input (fine-to-coarse)
        fine_to_coarse = outputs[::-1]
        return fine_to_coarse



from collections import OrderedDict
from functools import partial
from typing import Union

from torch import nn

from .common import EncoderModule, make_n_channel_input
from ..unet import UnetBlock

__all__ = ["UnetEncoder"]


class UnetEncoder(EncoderModule):
    """
    Vanilla U-Net encoder
    """

    def __init__(
        self,
        in_channels=3,
        out_channels=32,
        num_layers=4,
        growth_factor=2,
        pool_block: Union[nn.MaxPool2d, nn.AvgPool2d] = None,
        unet_block: Union[nn.Module, UnetBlock] = UnetBlock,
    ):
        if pool_block is None:
            pool_block = partial(nn.MaxPool2d, kernel_size=2, stride=2)

        feature_maps = [out_channels * (growth_factor**i) for i in range(num_layers)]
        strides = [2**i for i in range(num_layers)]
        super().__init__(feature_maps, strides, layers=list(range(num_layers)))

        input_filters = in_channels
        self.num_layers = num_layers
        for layer in range(num_layers):
            block = unet_block(input_filters, feature_maps[layer])

            if layer > 0:
                pool = pool_block()
                block = nn.Sequential(OrderedDict([("pool", pool), ("conv", block)]))

            input_filters = feature_maps[layer]
            self.add_module(f"layer{layer}", block)

    @property
    def encoder_layers(self):
        return [self.__getattr__(f"layer{layer}") for layer in range(self.num_layers)]

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.layer0.conv1 = make_n_channel_input(self.layer0.conv1, input_channels, mode)
        return self



from .common import EncoderModule, _take, make_n_channel_input
from ..backbone.inceptionv4 import inceptionv4

__all__ = ["InceptionV4Encoder"]


class InceptionV4Encoder(EncoderModule):
    def __init__(self, pretrained=True, layers=None, **kwargs):
        backbone = inceptionv4(pretrained="imagenet" if pretrained else None)
        channels = [64, 192, 384, 1024, 1536]
        strides = [2, 4, 8, 16, 32]  # Note output strides are approximate
        if layers is None:
            layers = [1, 2, 3, 4]
        features = backbone.features
        super().__init__(channels, strides, layers)

        self.layer0 = features[0:3]
        self.layer1 = features[3:5]
        self.layer2 = features[5:10]
        self.layer3 = features[10:18]
        self.layer4 = features[18:22]

        self._output_strides = _take(strides, layers)
        self._output_filters = _take(channels, layers)

    def forward(self, x):
        output_features = []
        for layer in self.encoder_layers:
            output = layer(x)
            output_features.append(output)
            x = output

        # Return only features that were requested
        return _take(output_features, self._layers)

    @property
    def encoder_layers(self):
        return [self.layer0, self.layer1, self.layer2, self.layer3, self.layer4]

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.layer0[0].conv = make_n_channel_input(self.layer0[0].conv, input_channels, mode)
        return self



import os.path
from pathlib import Path

from pytorch_toolbelt.utils.fs import read_rgb_image, read_image_as_is, change_extension

lena_str_name = os.path.join(os.path.dirname(__file__), "lena.png")
lena_path = Path(os.path.dirname(__file__)) / "lena.png"
lena_path_jpeg = Path(os.path.dirname(__file__)) / "lena.jpeg"


def test_read_rgb_image_str():
    img = read_rgb_image(lena_str_name)
    assert img.shape == (220, 220, 3)


def test_read_rgb_image_pathlib():
    img = read_rgb_image(lena_path)
    assert img.shape == (220, 220, 3)


def test_read_image_as_is_str():
    img = read_image_as_is(lena_str_name)
    assert img.shape == (220, 220, 3)


def test_read_image_as_is_pathlib():
    img = read_image_as_is(lena_path)
    assert img.shape == (220, 220, 3)


def test_change_extension_str():
    assert change_extension("lena.png", ".jpeg") == "lena.jpeg"


def test_change_extension_pathlib():
    assert change_extension(lena_path, "jpeg") == lena_path_jpeg
    assert change_extension(lena_path, ".jpeg") == lena_path_jpeg



import copy
import math
import warnings

import torch
from torch import nn

from .common import GenericTimmEncoder
from ..common import make_n_channel_input
from ...activations import ACT_SILU, get_activation_block

__all__ = ["TimmEfficientNetV2"]


def make_n_channel_input_conv2d_same(conv: nn.Conv2d, in_channels: int, mode="auto", **kwargs):
    assert isinstance(conv, nn.Conv2d)
    if conv.in_channels == in_channels:
        warnings.warn("make_n_channel_input call is spurious")
        return conv

    from timm.models.layers import Conv2dSame

    new_conv = Conv2dSame(
        in_channels,
        out_channels=conv.out_channels,
        kernel_size=kwargs.get("kernel_size", conv.kernel_size),
        stride=kwargs.get("stride", conv.stride),
        padding=kwargs.get("padding", conv.padding),
        dilation=kwargs.get("dilation", conv.dilation),
        groups=kwargs.get("groups", conv.groups),
        bias=kwargs.get("bias", conv.bias is not None),
    )

    w = conv.weight
    if in_channels > conv.in_channels:
        n = math.ceil(in_channels / float(conv.in_channels))
        w = torch.cat([w] * n, dim=1)
        w = w[:, :in_channels, ...]
        new_conv.weight = nn.Parameter(w, requires_grad=True)
    else:
        w = w[:, 0:in_channels, ...]
        new_conv.weight = nn.Parameter(w, requires_grad=True)

    return new_conv


class TimmEfficientNetV2(GenericTimmEncoder):
    def __init__(
        self,
        model_name: str = "efficientnetv2_rw_s",
        pretrained=True,
        layers=None,
        activation: str = ACT_SILU,
        drop_rate=0.0,
        drop_path_rate=0.0,
        first_conv_stride_one: bool = False,
    ):
        from timm.models.factory import create_model

        act_layer = get_activation_block(activation)
        encoder = create_model(
            model_name=model_name,
            pretrained=pretrained,
            features_only=True,
            act_layer=act_layer,
            drop_rate=drop_rate,
            drop_path_rate=drop_path_rate,
        )
        if first_conv_stride_one:
            encoder.conv_stem.stride = (1, 1)
            encoder.feature_info = copy.deepcopy(encoder.feature_info)
            for fi in encoder.feature_info:
                fi["reduction"] = fi["reduction"] // 2

        super().__init__(encoder, layers)

    @torch.jit.unused
    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        from timm.models.layers import Conv2dSame

        if isinstance(self.encoder.conv_stem, Conv2dSame):
            self.encoder.conv_stem = make_n_channel_input_conv2d_same(
                self.encoder.conv_stem, input_channels, mode, **kwargs
            )
        else:
            self.encoder.conv_stem = make_n_channel_input(self.encoder.conv_stem, input_channels, mode, **kwargs)
        return self



import itertools
from collections.abc import Sized, Iterable
from typing import Union, Tuple

import torch
from torch import Tensor

from ..utils.support import pytorch_toolbelt_deprecated

__all__ = [
    "geometric_mean",
    "harmonic_mean",
    "harmonic1p_mean",
    "logodd_mean",
    "log1p_mean",
    "pad_image_tensor",
    "pad_tensor_to_size",
    "torch_fliplr",
    "torch_flipud",
    "torch_none",
    "torch_rot180",
    "torch_rot270",
    "torch_rot90",
    "torch_rot90_ccw",
    "torch_rot90_ccw_transpose",
    "torch_rot90_cw",
    "torch_rot90_cw_transpose",
    "torch_transpose",
    "torch_transpose2",
    "torch_transpose_",
    "torch_transpose_rot90_ccw",
    "torch_transpose_rot90_cw",
    "unpad_image_tensor",
    "unpad_xyxy_bboxes",
]


def torch_none(x: Tensor) -> Tensor:
    """
    Return input argument without any modifications
    :param x: input tensor
    :return: x
    """
    return x


def torch_rot90_ccw(x):
    return x.rot90(k=1, dims=(2, 3))


def torch_rot90_cw(x):
    return x.rot90(k=-1, dims=(2, 3))


def torch_rot90_ccw_transpose(x):
    return x.rot90(k=1, dims=(2, 3)).transpose(2, 3)


def torch_rot90_cw_transpose(x):
    return x.rot90(k=-1, dims=(2, 3)).transpose(2, 3)


def torch_transpose_rot90_ccw(x: Tensor):
    return x.transpose(2, 3).rot90(k=1, dims=(2, 3))


def torch_transpose_rot90_cw(x):
    return x.transpose(2, 3).rot90(k=-1, dims=(2, 3))


@pytorch_toolbelt_deprecated("Function torch_rot90 has been marked as deprecated. Please use torch_rot90_ccw instead")
def torch_rot90(x: Tensor):
    """
    Rotate 4D image tensor by 90 degrees
    :param x:
    :return:
    """
    return torch_rot90_ccw(x)


def torch_rot180(x: Tensor):
    """
    Rotate 4D image tensor by 180 degrees
    :param x:
    :return:
    """
    return torch.rot90(x, k=2, dims=(2, 3))


def torch_rot180_transpose(x):
    return x.rot90(k=2, dims=(2, 3)).transpose(2, 3)


def torch_transpose_rot180(x):
    return x.transpose(2, 3).rot90(k=2, dims=(2, 3))


@pytorch_toolbelt_deprecated("Function torch_rot270 has been marked as deprecated. Please use torch_rot90_cw instead")
def torch_rot270(x: Tensor):
    """
    Rotate 4D image tensor by 270 degrees
    :param x:
    :return:
    """
    return torch_rot90_cw(x)


def torch_flipud(x: Tensor):
    """
    Flip 4D image tensor vertically
    :param x:
    :return:
    """
    return x.flip(2)


def torch_fliplr(x: Tensor):
    """
    Flip 4D image tensor horizontally
    :param x:
    :return:
    """
    return x.flip(3)


def torch_transpose(x: Tensor):
    """
    Transpose 4D image tensor by main image diagonal
    :param x:
    :return:
    """
    return x.transpose(2, 3)


def torch_transpose_(x: Tensor):
    return x.transpose_(2, 3)


def torch_transpose2(x: Tensor):
    """
    Transpose 4D image tensor by second image diagonal
    :param x:
    :return:
    """
    return x.transpose(3, 2)


def pad_tensor_to_size(x: Tensor, size: Tuple[int, ...], mode="constant", value=0) -> Tuple[Tensor, Tuple[slice, ...]]:
    """
    Pad tensor to given size by appending elements to the beginning and end of each axis.

    :param x: Input tensor of shape [B, C, *num_spatial_dims]
    :param size: Target tensor size defined as an array of [num_spatial_dims] elements
    :param mode: Padding mode, see torch.nn.functional.pad
    :param value: Padding value, see torch.nn.functional.pad
    :return: Tuple of padded tensor and crop parameters. Second argument can be used to reverse pad operation of model output
    """
    num_spatial_dims = len(size)
    if num_spatial_dims != len(x.shape) - 2:
        raise ValueError(f"Expected {num_spatial_dims} spatial dimensions, got {len(x.shape) - 2}")

    spatial_dims = x.shape[-num_spatial_dims:]
    padding = torch.tensor(size) - torch.tensor(spatial_dims)
    padding_before = padding // 2
    padding_after = padding - padding_before

    padding_pairs = tuple(zip(padding_before.tolist(), padding_after.tolist()))
    padding_params = tuple(itertools.chain(*reversed(padding_pairs)))

    x = torch.nn.functional.pad(x, pad=padding_params, mode=mode, value=value)

    crop_params = [slice(None), slice(None)] + [
        slice(before, before + total_size)
        for (before, after, total_size) in zip(padding_before, padding_after, spatial_dims)
    ]
    return x, crop_params


def pad_image_tensor(
    image_tensor: Tensor, pad_size: Union[int, Tuple[int, int]] = 32
) -> Tuple[Tensor, Tuple[int, int, int, int]]:
    """Pad input tensor to make it's height and width dividable by @pad_size

    :param image_tensor: 4D image tensor of shape NCHW
    :param pad_size: Pad size
    :return: Tuple of output tensor and pad params. Second argument can be used to reverse pad operation of model output
    """
    if len(image_tensor.size()) != 4:
        raise ValueError("Tensor must have rank 4 ([B,C,H,W])")

    rows, cols = image_tensor.size(2), image_tensor.size(3)
    if isinstance(pad_size, Sized) and isinstance(pad_size, Iterable) and len(pad_size) == 2:
        pad_height, pad_width = [int(val) for val in pad_size]
    elif isinstance(pad_size, int):
        pad_height = pad_width = pad_size
    else:
        raise ValueError(
            f"Unsupported pad_size: {pad_size}, must be either tuple(pad_rows,pad_cols) or single int scalar."
        )

    if rows > pad_height:
        pad_rows = rows % pad_height
        pad_rows = pad_height - pad_rows if pad_rows > 0 else 0
    else:
        pad_rows = pad_height - rows

    if cols > pad_width:
        pad_cols = cols % pad_width
        pad_cols = pad_width - pad_cols if pad_cols > 0 else 0
    else:
        pad_cols = pad_width - cols

    if pad_rows == 0 and pad_cols == 0:
        return image_tensor, (0, 0, 0, 0)

    pad_top = pad_rows // 2
    pad_btm = pad_rows - pad_top

    pad_left = pad_cols // 2
    pad_right = pad_cols - pad_left

    pad = [pad_left, pad_right, pad_top, pad_btm]
    image_tensor = torch.nn.functional.pad(image_tensor, pad)
    return image_tensor, pad


def unpad_image_tensor(image_tensor: Tensor, pad) -> Tensor:
    if len(image_tensor.size()) != 4:
        raise ValueError("Tensor must have rank 4 ([B,C,H,W])")

    pad_left, pad_right, pad_top, pad_btm = pad
    rows, cols = image_tensor.size(2), image_tensor.size(3)
    return image_tensor[..., pad_top : rows - pad_btm, pad_left : cols - pad_right]


def unpad_xyxy_bboxes(bboxes_tensor: torch.Tensor, pad, dim=-1):
    pad_left, pad_right, pad_top, pad_btm = pad
    pad = torch.tensor([pad_left, pad_top, pad_left, pad_top], dtype=bboxes_tensor.dtype).to(bboxes_tensor.device)

    if dim == -1:
        dim = len(bboxes_tensor.size()) - 1

    expand_dims = list(set(range(len(bboxes_tensor.size()))) - {dim})
    for i, dim in enumerate(expand_dims):
        pad = pad.unsqueeze(dim)

    return bboxes_tensor - pad


def geometric_mean(x: Tensor, dim: int) -> Tensor:
    """
    Compute geometric mean along given dimension.
    This implementation assume values are in range (0...1) (Probabilities)
    Args:
        x: Input tensor of arbitrary shape
        dim: Dimension to reduce

    Returns:
        Tensor
    """
    return x.log().mean(dim=dim).exp()


def harmonic_mean(x: Tensor, dim: int, eps: float = 1e-6) -> Tensor:
    """
    Compute harmonic mean along given dimension.

    Args:
        x: Input tensor of arbitrary shape
        dim: Dimension to reduce

    Returns:
        Tensor
    """
    x = torch.reciprocal(x.clamp_min(eps))
    x = torch.mean(x, dim=dim)
    x = torch.reciprocal(x.clamp_min(eps))
    return x


def harmonic1p_mean(x: Tensor, dim: int) -> Tensor:
    """
    Compute harmonic mean along given dimension.

    Args:
        x: Input tensor of arbitrary shape
        dim: Dimension to reduce

    Returns:
        Tensor
    """
    x = torch.reciprocal(x + 1)
    x = torch.mean(x, dim=dim)
    x = torch.reciprocal(x) - 1
    return x


def logodd_mean(x: Tensor, dim: int, eps: float = 1e-6) -> Tensor:
    """
    Compute log-odd mean along given dimension.
    logodd = log(p / (1 - p))

    This implementation assume values are in range [0, 1] (Probabilities)
    Args:
        x: Input tensor of arbitrary shape
        dim: Dimension to reduce

    Returns:
        Tensor
    """
    x = x.clamp(min=eps, max=1.0 - eps)
    x = torch.log(x / (1 - x))
    x = torch.mean(x, dim=dim)
    x = torch.exp(x) / (1 + torch.exp(x))
    return x


def log1p_mean(x: Tensor, dim: int) -> Tensor:
    """
    Compute average log(x+1) and them compute exp.
    Requires all inputs to be non-negative

    Args:
        x: Input tensor of arbitrary shape
        dim: Dimension to reduce

    Returns:
        Tensor
    """
    x = torch.log1p(x)
    x = torch.mean(x, dim=dim)
    x = torch.exp(x) - 1
    return x



import torch
from torch import nn


class SRMLayer(nn.Module):
    """An implementation of SRM block, proposed in
    "SRM : A Style-based Recalibration Module for Convolutional Neural Networks".

    """

    def __init__(self, channels: int):
        super(SRMLayer, self).__init__()

        # Equal to torch.einsum('bck,ck->bc', A, B)
        self.cfc = nn.Conv1d(channels, channels, kernel_size=2, bias=False, groups=channels)
        self.bn = nn.BatchNorm1d(channels)

    def forward(self, x):
        b, c, _, _ = x.size()

        # Style pooling
        mean = x.view(b, c, -1).mean(-1).unsqueeze(-1)
        std = x.view(b, c, -1).std(-1).unsqueeze(-1)
        u = torch.cat((mean, std), -1)  # (b, c, 2)

        # Style integration
        z = self.cfc(u)  # (b, c, 1)
        z = self.bn(z)
        g = torch.sigmoid(z)
        g = g.view(b, c, 1, 1)

        return x * g.expand_as(x)



from collections import namedtuple
from typing import Optional

import numpy as np
import torch
from pytorch_toolbelt.utils import to_numpy
from torchvision.ops import box_iou

__all__ = ["match_bboxes", "match_bboxes_hungarian", "BBoxesMatchResult"]

BBoxesMatchResult = namedtuple(
    "BBoxesMatchResult",
    [
        # Array of shape [num_classes]
        "true_positives",
        # Array of shape [num_classes]
        "false_positives",
        # Array of shape [num_classes]
        "false_negatives",
        # Matrix of shape [num_classes+1, num_classes+1], where last class corresponds to None,
        # in other words - no detection. The notation is confusion_matrix[gt, pred].
        "confusion_matrix",
        # An array of [K,2], where K is number of true positive matches and second dimension contains
        # tuple of (pred_boxes_index, true_boxes_index) of true positive matches
        "true_positive_indexes",
    ],
)


@torch.no_grad()
def match_bboxes(
    pred_boxes: np.ndarray,
    pred_labels: np.ndarray,
    pred_scores: np.ndarray,
    true_boxes: np.ndarray,
    true_labels: np.ndarray,
    num_classes: int,
    iou_threshold: float = 0.5,
) -> BBoxesMatchResult:
    """
    Match predictect and ground-truth bounding boxes. Boxes with higher confidence are matched the first.
    There can be only one match between predicted and ground-truth box.

    For multi-class case, if the boxes match, but their classes does not match, this counts as 1 FN
    to ground-truth class and 1 FP to predicted class.

    :param pred_boxes: Detected bboxes in [x1, y1, x2, y2] format of shape [N,4]
    :param pred_labels: Detected labels of shape [N]
    :param pred_scores: Detected scores of shape [N]. Optional
    :param true_boxes:  Ground-truth bboxes in [x1, y1, x2, y2] format of shape [M,4]
    :param true_labels: Ground-truth labels of shape [M]
    :param num_classes: Total number of classes
    :param iou_threshold: IoU threshold to count detection as "match"
    :return:
        Tuple of [num_classes], [num_classes], [num_classes] corresponding to
        true positives, false positive and false negative counts per class
    """
    if len(pred_labels) != len(pred_boxes) or len(pred_labels) != len(pred_scores):
        raise ValueError(
            f"Inconsistent lengths of predicted bboxes:{len(pred_boxes)} labels:{len(pred_labels)} and their scores: {len(pred_scores)}"
        )

    if len(true_boxes) != len(true_labels):
        raise ValueError(
            f"Inconsistent lengths of ground-truth bboxes:{len(true_boxes)} and their labels:{len(true_labels)}"
        )

    true_positives = np.zeros(num_classes, dtype=int)
    false_positives = np.zeros(num_classes, dtype=int)
    false_negatives = np.zeros(num_classes, dtype=int)

    # Confusion matrix [gt, pred]
    confusion_matrix = np.zeros((num_classes + 1, num_classes + 1), dtype=int)
    none_class = num_classes

    num_pred_objects = len(pred_boxes)
    num_true_objects = len(true_boxes)

    if num_pred_objects == 0 and num_true_objects == 0:
        return BBoxesMatchResult(
            true_positives=true_positives,
            false_positives=false_positives,
            false_negatives=false_negatives,
            confusion_matrix=confusion_matrix,
            true_positive_indexes=np.zeros((0, 2), dtype=int),
        )
    elif num_pred_objects == 0:
        for true_class in true_labels:
            false_negatives[true_class] += 1
            confusion_matrix[true_class, none_class] += 1
        return BBoxesMatchResult(
            true_positives=true_positives,
            false_positives=false_positives,
            false_negatives=false_negatives,
            confusion_matrix=confusion_matrix,
            true_positive_indexes=np.zeros((0, 2), dtype=int),
        )
    elif num_true_objects == 0:
        for pred_class in pred_labels:
            false_positives[pred_class] += 1
            confusion_matrix[none_class, pred_class] += 1
        return BBoxesMatchResult(
            true_positives=true_positives,
            false_positives=false_positives,
            false_negatives=false_negatives,
            confusion_matrix=confusion_matrix,
            true_positive_indexes=np.zeros((0, 2), dtype=int),
        )

    # Reorder predictions to start matching with the most confident ones
    order = np.argsort(-pred_scores)
    rorder = np.argsort(order)
    pred_boxes = pred_boxes[order]
    pred_labels = pred_labels[order]
    #
    iou_matrix: np.ndarray = to_numpy(
        box_iou(torch.from_numpy(pred_boxes).float(), torch.from_numpy(true_boxes).float())
    )

    remainig_preds = np.ones(num_pred_objects, dtype=bool)
    remainig_trues = np.ones(num_true_objects, dtype=bool)
    true_positive_indexes = []

    for ci in range(num_true_objects):
        # Find a first prediction box with IoU greater than or equal iou threshold with a groundtruth box
        candidates = np.flatnonzero(iou_matrix[:, ci] >= iou_threshold)
        if len(candidates):
            ri = candidates[0]

            iou_matrix[ri, :] = 0

            remainig_preds[ri] = False
            remainig_trues[ci] = False

            pred_class = pred_labels[ri]
            true_class = true_labels[ci]

            if pred_class == true_class:
                # If there is a matching bbox found above, increase the count of true positives by one (TP).
                true_positives[true_class] += 1
                # Inverse indexe of predicted boxes here to keep the order of original (unsorted) bboxes
                true_positive_indexes.append((rorder[ri], ci))
            else:
                # If classes does not match, then we add false-positive for predicted class and
                # false-negative to target class
                false_positives[pred_class] += 1
                false_negatives[true_class] += 1

            confusion_matrix[true_class, pred_class] += 1

    if remainig_preds.any():
        for pred_class in pred_labels[remainig_preds]:
            false_positives[pred_class] += 1
            confusion_matrix[none_class, pred_class] += 1

    if remainig_trues.any():
        for true_class in true_labels[remainig_trues]:
            false_negatives[true_class] += 1
            confusion_matrix[true_class, none_class] += 1

    return BBoxesMatchResult(
        true_positives=true_positives,
        true_positive_indexes=np.array(true_positive_indexes, dtype=int).reshape((-1, 2)),
        false_positives=false_positives,
        false_negatives=false_negatives,
        confusion_matrix=confusion_matrix,
    )


@torch.no_grad()
def match_bboxes_hungarian(
    pred_boxes: np.ndarray,
    pred_labels: np.ndarray,
    true_boxes: np.ndarray,
    true_labels: np.ndarray,
    num_classes: int,
    iou_threshold: float = 0.5,
) -> BBoxesMatchResult:
    """
    Match predictect and ground-truth bounding boxes using hungarian matching algorithm.

    For multi-class case, if the boxes match, but their classes does not match, this counts as 1 FN
    to ground-truth class and 1 FP to predicted class.

    :param pred_boxes: Detected bboxes in [x1, y1, x2, y2] format of shape [N,4]
    :param pred_labels: Detected labels of shape [N]
    :param true_boxes:  Ground-truth bboxes in [x1, y1, x2, y2] format of shape [M,4]
    :param true_labels: Ground-truth labels of shape [M]
    :param num_classes: Total number of classes
    :param iou_threshold: IoU threshold to count detection as "match"
    :param min_size: If not None, will exclude boxes with area smaller than this parameter from evaluation
    :return:
        Tuple of [num_classes], [num_classes], [num_classes] corresponding to
        true positives, false positive and false negative counts per class
    """
    from scipy.optimize import linear_sum_assignment

    if len(pred_labels) != len(pred_boxes):
        raise ValueError(f"Inconsistent lengths of predicted bboxes:{len(pred_boxes)} labels:{len(pred_labels)}")

    if len(true_boxes) != len(true_labels):
        raise ValueError(
            f"Inconsistent lengths of ground-truth bboxes:{len(true_boxes)} and their labels:{len(true_labels)}"
        )

    true_positives = np.zeros(num_classes, dtype=int)
    false_positives = np.zeros(num_classes, dtype=int)
    false_negatives = np.zeros(num_classes, dtype=int)

    # Confusion matrix [gt, pred]
    confusion_matrix = np.zeros((num_classes + 1, num_classes + 1), dtype=int)
    none_class = num_classes

    num_pred_objects = len(pred_boxes)
    num_true_objects = len(true_boxes)

    if num_pred_objects == 0 and num_true_objects == 0:
        return BBoxesMatchResult(
            true_positives=true_positives,
            false_positives=false_positives,
            false_negatives=false_negatives,
            confusion_matrix=confusion_matrix,
            true_positive_indexes=np.zeros((0, 2), dtype=int),
        )
    elif num_pred_objects == 0:
        for true_class in true_labels:
            false_negatives[true_class] += 1
            confusion_matrix[true_class, none_class] += 1
        return BBoxesMatchResult(
            true_positives=true_positives,
            false_positives=false_positives,
            false_negatives=false_negatives,
            confusion_matrix=confusion_matrix,
            true_positive_indexes=np.zeros((0, 2), dtype=int),
        )
    elif num_true_objects == 0:
        for pred_class in pred_labels:
            false_positives[pred_class] += 1
            confusion_matrix[none_class, pred_class] += 1
        return BBoxesMatchResult(
            true_positives=true_positives,
            false_positives=false_positives,
            false_negatives=false_negatives,
            confusion_matrix=confusion_matrix,
            true_positive_indexes=np.zeros((0, 2), dtype=int),
        )

    iou_matrix = to_numpy(box_iou(torch.from_numpy(pred_boxes).float(), torch.from_numpy(true_boxes).float()))
    row_ind, col_ind = linear_sum_assignment(iou_matrix, maximize=True)

    remainig_preds = np.ones(num_pred_objects, dtype=bool)
    remainig_trues = np.ones(num_true_objects, dtype=bool)

    true_positive_indexes = []

    for ri, ci in zip(row_ind, col_ind):
        pred_class = pred_labels[ri]
        true_class = true_labels[ci]
        if iou_matrix[ri, ci] >= iou_threshold:
            remainig_preds[ri] = False
            remainig_trues[ci] = False
            if pred_class == true_class:
                # If there is a matching polygon found above, increase the count of true positives by one (TP).
                true_positives[true_class] += 1
                true_positive_indexes.append((ri, ci))
            else:
                # If classes does not match, then we add false-positive for predicted class and
                # false-negative to target class
                false_positives[pred_class] += 1
                false_negatives[true_class] += 1

            confusion_matrix[true_class, pred_class] += 1

    if remainig_preds.any():
        for pred_class in pred_labels[remainig_preds]:
            false_positives[pred_class] += 1
            confusion_matrix[none_class, pred_class] += 1

    if remainig_trues.any():
        for true_class in true_labels[remainig_trues]:
            false_negatives[true_class] += 1
            confusion_matrix[true_class, none_class] += 1

    return BBoxesMatchResult(
        true_positives=true_positives,
        true_positive_indexes=np.array(true_positive_indexes, dtype=int).reshape((-1, 2)),
        false_positives=false_positives,
        false_negatives=false_negatives,
        confusion_matrix=confusion_matrix,
    )



from __future__ import absolute_import

import itertools
import math
import warnings
from typing import List, Iterable, Tuple, Optional

import cv2
import numpy as np

from .torch_utils import image_to_tensor

__all__ = [
    "plot_confusion_matrix",
    "plot_compressed_confusion_matrix",
    "render_figure_to_tensor",
    "hstack_autopad",
    "vstack_autopad",
    "vstack_header",
    "grid_stack",
    "plot_heatmap",
]


def plot_heatmap(
    cm: np.ndarray,
    title: str,
    x_label=None,
    y_label=None,
    x_ticks: List[str] = None,
    y_ticks: List[str] = None,
    format_string=None,
    show_scores=True,
    fontsize=12,
    figsize: Tuple[int, int] = (16, 16),
    fname=None,
    noshow: bool = False,
    cmap=None,
    backend="Agg",
):
    if len(cm.shape) != 2:
        raise ValueError("Heatmap must be a 2-D array")
    import matplotlib

    matplotlib.use(backend)
    import matplotlib.pyplot as plt

    if cmap is None:
        cmap = plt.cm.Oranges

    f = plt.figure(figsize=figsize)
    plt.imshow(cm, interpolation="nearest", cmap=cmap)
    plt.title(title)
    plt.colorbar(fraction=0.046, pad=0.04)

    if x_ticks is not None:
        plt.xticks(np.arange(len(x_ticks)), x_ticks, rotation=45, ha="right")

    if y_ticks is not None:
        plt.yticks(np.arange(len(y_ticks)), y_ticks)

    if format_string is None:
        format_string = ".2f" if np.issubdtype(cm.dtype, np.floating) else "d"

    if show_scores:
        thresh = (cm.max() + cm.min()) / 2.0
        for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
            text = format(cm[i, j], format_string) if np.isfinite(cm[i, j]) else "N/A"
            color = "white" if cm[i, j] > thresh else "black"
            plt.text(
                j,
                i,
                text,
                horizontalalignment="center",
                verticalalignment="center_baseline",
                fontsize=fontsize,
                color=color,
            )

    plt.xlabel(x_label)
    plt.ylabel(y_label)

    plt.tight_layout()

    if fname is not None:
        plt.savefig(fname=fname, dpi=200)

    if not noshow:
        plt.show()

    return f


def plot_compressed_confusion_matrix(
    cm: np.ndarray,
    figsize: Tuple[int, int] = (16, 16),
    normalize: bool = False,
    title: str = "Confusion matrix",
    cmap=None,
    fname=None,
    noshow: bool = False,
):
    from matplotlib import pyplot as plt

    if normalize:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cm = cm.astype(np.float32) / cm.sum(axis=1)[:, np.newaxis]
    else:
        accuracy = np.trace(cm) / (float(np.sum(cm)) + 1e-8)
        misclass = 1 - accuracy

    f = plt.figure(figsize=figsize)
    plt.imshow(cm, interpolation="nearest", cmap=cmap)
    plt.colorbar(fraction=0.046, pad=0.04)
    plt.title(title)

    if normalize:
        # We don't show Accuracy & Misclassification scores for normalized CM
        plt.xlabel("Predicted label")
    else:
        plt.xlabel("Predicted label\nAccuracy={:0.4f}; Misclass={:0.4f}".format(accuracy, misclass))

    plt.xlabel("Predicted label")
    plt.ylabel("True label")
    plt.tight_layout()
    if fname is not None:
        plt.savefig(fname=fname, dpi=200)
    if not noshow:
        plt.show()
    return f


def plot_confusion_matrix(
    cm: np.ndarray,
    class_names: List[str],
    figsize: Tuple[int, int] = (16, 16),
    fontsize: int = 12,
    normalize: bool = False,
    title: str = "Confusion matrix",
    cmap=None,
    fname=None,
    show_scores: bool = True,
    noshow: bool = False,
    backend: str = "Agg",
    format_string: Optional[str] = None,
):
    """
    Render the confusion matrix and return matplotlib's figure with it.
    Normalization can be applied by setting `normalize=True`.

    Args:
        cm: Numpy array of (N,N) shape - confusion matrix array
        class_names: List of [N] names of the classes
        figsize:
        fontsize:
        normalize: Whether to apply normalization for each row of CM
        title: Title of the confusion matrix
        cmap:
        fname: Filename of the rendered confusion matrix
        show_scores: Show scores in each cell
        noshow:
        backend:
        format_string:

    Returns:
        Matplotlib's figure
    """
    import matplotlib

    matplotlib.use(backend)
    import matplotlib.pyplot as plt

    if cmap is None:
        cmap = plt.cm.Oranges

    if normalize:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cm = cm.astype(np.float32) / cm.sum(axis=1)[:, np.newaxis]
    else:
        accuracy = np.trace(cm) / (float(np.sum(cm)) + 1e-8)
        misclass = 1 - accuracy

    f = plt.figure(figsize=figsize)
    plt.imshow(cm, interpolation="nearest", cmap=cmap)
    plt.title(title)
    plt.colorbar(fraction=0.046, pad=0.04)

    tick_marks = np.arange(len(class_names))
    plt.xticks(tick_marks, class_names, rotation=45, ha="right")
    plt.yticks(tick_marks, class_names)

    if format_string is None:
        format_string = ".3f" if normalize else "d"

    if show_scores:
        thresh = (cm.max() + cm.min()) / 2.0
        for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
            text = format(cm[i, j], format_string) if np.isfinite(cm[i, j]) else "N/A"
            color = "white" if cm[i, j] > thresh else "black"
            plt.text(j, i, text, horizontalalignment="center", fontsize=fontsize, color=color)

    plt.ylabel("True label")

    if normalize:
        # We don't show Accuracy & Misclassification scores for normalized CM
        plt.xlabel("Predicted label")
    else:
        plt.xlabel("Predicted label\nAccuracy={:0.4f}; Misclass={:0.4f}".format(accuracy, misclass))

    plt.tight_layout()

    if fname is not None:
        plt.savefig(fname=fname, dpi=200)

    if not noshow:
        plt.show()

    return f


def render_figure_to_tensor(figure):
    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot as plt

    figure.canvas.draw()

    # string = figure.canvas.tostring_argb()

    image = np.array(figure.canvas.renderer._renderer)
    plt.close(figure)
    del figure

    image = image_to_tensor(image)
    return image


def hstack_autopad(images: Iterable[np.ndarray], pad_value: int = 0, spacing=0) -> np.ndarray:
    """
    Stack images horizontally with automatic padding

    Args:
        images: List of images to stack

    Returns:
        image
    """
    max_height = 0
    for img in images:
        max_height = max(max_height, img.shape[0])

    padded_images = []
    for img_index, img in enumerate(images):
        is_last_image = img_index == len(images) - 1
        height = img.shape[0]
        pad_top = 0
        pad_bottom = max_height - height
        pad_left = 0
        pad_right = 0 if is_last_image else spacing
        img = cv2.copyMakeBorder(img, pad_top, pad_bottom, pad_left, pad_right, cv2.BORDER_CONSTANT, value=pad_value)
        (rows, cols) = img.shape[0:2]
        padded_images.append(img)

    return np.hstack(padded_images)


def vstack_autopad(images: Iterable[np.ndarray], pad_value: int = 0, spacing: int = 0) -> np.ndarray:
    """
    Stack images vertically with automatic padding

    Args:
        images: List of images to stack

    Returns:
        image
    """
    max_width = 0
    for img in images:
        max_width = max(max_width, img.shape[1])

    padded_images = []
    for img_index, img in enumerate(images):
        is_last_image = img_index == len(images) - 1
        width = img.shape[1]
        pad_top = 0
        pad_bottom = 0 if is_last_image else spacing
        pad_left = 0
        pad_right = max_width - width
        img = cv2.copyMakeBorder(img, pad_top, pad_bottom, pad_left, pad_right, cv2.BORDER_CONSTANT, value=pad_value)
        padded_images.append(img)

    return np.vstack(padded_images)


def vstack_header(
    image: np.ndarray,
    title: str,
    bg_color=(35, 41, 40),
    text_color=(242, 248, 248),
    text_thickness: int = 2,
    text_scale=1.5,
) -> np.ndarray:
    (rows, cols) = image.shape[:2]

    title_image = np.zeros((30, cols, 3), dtype=np.uint8)
    title_image[:] = bg_color
    cv2.putText(
        title_image,
        title,
        (10, 24),
        fontFace=cv2.FONT_HERSHEY_PLAIN,
        fontScale=text_scale,
        color=text_color,
        thickness=text_thickness,
        lineType=cv2.LINE_AA,
    )

    return vstack_autopad([title_image, image])


def grid_stack(
    images: List[np.ndarray], rows: int = None, cols: int = None, bg_color=0, spacing: int = 0
) -> np.ndarray:
    if rows is None and cols is None:
        rows = int(math.ceil(math.sqrt(len(images))))
        cols = int(math.ceil(len(images) / rows))
    elif rows is None:
        rows = math.ceil(len(images) / cols)
    elif cols is None:
        cols = math.ceil(len(images) / rows)
    else:
        if len(images) > rows * cols:
            raise ValueError("Number of rows * cols must be greater than number of images")

    image_rows = []
    for r in range(rows):
        image_rows.append(hstack_autopad(images[r * cols : (r + 1) * cols], bg_color=bg_color, spacing=spacing))

    return vstack_autopad(image_rows, bg_color=bg_color, spacing=spacing)



from typing import List

import torch
from torch import nn, Tensor

__all__ = ["DeeplabV3Decoder"]

from ..activations import instantiate_activation_block, ACT_RELU
from ..interfaces import AbstractDecoder, FeatureMapsSpecification
from ..spp import ASPPModule, ASPP


class DeeplabV3Decoder(AbstractDecoder):
    """
    Implements DeepLabV3 model from `"Rethinking Atrous Convolution for Semantic Image Segmentation"
    <https://arxiv.org/abs/1706.05587>`_.

    Partially copy-pasted from https://github.com/pytorch/vision/blob/master/torchvision/models/segmentation/deeplabv3.py
    """

    def __init__(
        self,
        input_spec: FeatureMapsSpecification,
        out_channels: int,
        aspp_channels: int,
        atrous_rates=(12, 24, 36),
        dropout: float = 0.5,
        activation=ACT_RELU,
    ):
        """

        Args:
            feature_maps: List of input channels
            aspp_channels:
            out_channels: Output channels
            atrous_rates:
            dropout:
            activation:
        """
        super().__init__(input_spec)
        self.aspp = ASPP(
            in_channels=input_spec.channels[-1],
            out_channels=aspp_channels,
            aspp_module=ASPPModule,
            atrous_rates=atrous_rates,
            dropout=dropout,
            activation=activation,
        )
        self.final = nn.Sequential(
            nn.Conv2d(aspp_channels, aspp_channels, 3, padding=1, bias=False),
            nn.BatchNorm2d(aspp_channels),
            instantiate_activation_block(activation, inplace=True),
            nn.Conv2d(aspp_channels, out_channels, kernel_size=1),
        )

        self.output_spec = FeatureMapsSpecification(channels=[out_channels], strides=[input_spec.strides[-1]])

    def forward(self, feature_maps: List[Tensor]) -> List[Tensor]:
        high_level_features = feature_maps[-1]
        high_level_features = self.aspp(high_level_features)
        return self.final(high_level_features)

    @torch.jit.unused
    def get_output_spec(self) -> FeatureMapsSpecification:
        return self.output_spec



from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn, Tensor

__all__ = ["BalancedBCEWithLogitsLoss", "balanced_binary_cross_entropy_with_logits"]


def balanced_binary_cross_entropy_with_logits(
    logits: Tensor, targets: Tensor, gamma: float = 1.0, ignore_index: Optional[int] = None, reduction: str = "mean"
) -> Tensor:
    """
    Balanced binary cross entropy loss.

    Args:
        logits:
        targets: This loss function expects target values to be hard targets 0/1.
        gamma: Power factor for balancing weights
        ignore_index:
        reduction:

    Returns:
        Zero-sized tensor with reduced loss if `reduction` is `sum` or `mean`; Otherwise returns loss of the
        shape of `logits` tensor.
    """
    pos_targets: Tensor = targets.eq(1).sum()
    neg_targets: Tensor = targets.eq(0).sum()

    num_targets = pos_targets + neg_targets
    pos_weight = torch.pow(neg_targets / (num_targets + 1e-7), gamma)
    neg_weight = 1.0 - pos_weight

    pos_term = pos_weight.pow(gamma) * targets * torch.nn.functional.logsigmoid(logits)
    neg_term = neg_weight.pow(gamma) * (1 - targets) * torch.nn.functional.logsigmoid(-logits)

    loss = -(pos_term + neg_term)

    if ignore_index is not None:
        loss = torch.masked_fill(loss, targets.eq(ignore_index), 0)

    if reduction == "mean":
        loss = loss.mean()

    if reduction == "sum":
        loss = loss.sum()

    return loss


class BalancedBCEWithLogitsLoss(nn.Module):
    """
    Balanced binary cross-entropy loss.

    https://arxiv.org/pdf/1504.06375.pdf (Formula 2)
    """

    __constants__ = ["gamma", "reduction", "ignore_index"]

    def __init__(self, gamma: float = 1.0, reduction="mean", ignore_index: Optional[int] = None):
        """

        Args:
            gamma:
            ignore_index:
            reduction:
        """
        super().__init__()
        self.gamma = gamma
        self.reduction = reduction
        self.ignore_index = ignore_index

    def forward(self, output: Tensor, target: Tensor) -> Tensor:
        return balanced_binary_cross_entropy_with_logits(
            output, target, gamma=self.gamma, ignore_index=self.ignore_index, reduction=self.reduction
        )



import torch
from torch import nn

from pytorch_toolbelt.modules.initialization import first_class_background_init


def test_first_class_background_init_conv():
    conv = nn.Conv2d(256, 19, kernel_size=3, padding=1)
    first_class_background_init(conv, background_prob=0.96)

    input = torch.randn(4, 256, 64, 64)
    output = conv(input).softmax(dim=1)

    actual_probas = output.mean(dim=(0, 2, 3))
    assert actual_probas[0] > 0.96
    assert (actual_probas[1:] < 0.04).all()


def test_first_class_background_init_linear():
    conv = nn.Linear(256, 19)
    first_class_background_init(conv, background_prob=0.96)

    input = torch.randn(4, 256)
    output = conv(input).softmax(dim=1)

    actual_probas = output.mean(dim=(0))
    assert actual_probas[0] > 0.96
    assert (actual_probas[1:] < 0.04).all()



from typing import List

import torch
import torch.nn.functional as F
from torch import nn, Tensor

from ..activations import instantiate_activation_block, ACT_RELU
from ..interfaces import AbstractDecoder, FeatureMapsSpecification

__all__ = ["DeeplabV3PlusDecoder"]

from ..spp import ASPP, SeparableASPPModule


class DeeplabV3PlusDecoder(AbstractDecoder):
    """
    Implements DeepLabV3 model from `"Rethinking Atrous Convolution for Semantic Image Segmentation"
    <https://arxiv.org/abs/1706.05587>`_.

    Partially copy-pasted from https://github.com/pytorch/vision/blob/master/torchvision/models/segmentation/deeplabv3.py
    """

    def __init__(
        self,
        input_spec: FeatureMapsSpecification,
        out_channels: int,
        aspp_channels: int,
        low_level_channels: int = 48,
        atrous_rates=(12, 24, 36),
        dropout: float = 0.5,
        activation: str = ACT_RELU,
    ):
        """

        Args:
            feature_maps: Input feature maps
            aspp_channels:
            out_channels: Number of output channels
            atrous_rates:
            dropout:
            activation:
            low_level_channels:
        """
        super().__init__(input_spec)

        self.project = nn.Sequential(
            nn.Conv2d(input_spec.channels[0], low_level_channels, 1, bias=False),
            nn.BatchNorm2d(low_level_channels),
            instantiate_activation_block(activation, inplace=True),
        )

        self.aspp = ASPP(
            in_channels=input_spec.channels[-1],
            out_channels=aspp_channels,
            atrous_rates=atrous_rates,
            dropout=dropout,
            activation=activation,
            aspp_module=SeparableASPPModule,
        )

        self.final = nn.Sequential(
            nn.Conv2d(
                aspp_channels + low_level_channels,
                out_channels,
                kernel_size=3,
                padding=1,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels),
            instantiate_activation_block(activation, inplace=True),
        )

        self.output_spec = FeatureMapsSpecification(
            channels=[out_channels, aspp_channels], strides=[input_spec.strides[0], input_spec.strides[-1]]
        )

    def forward(self, feature_maps: List[Tensor]) -> List[Tensor]:
        coarse_features = self.aspp(feature_maps[-1])
        low_level_features = self.project(feature_maps[0])

        coarse_features_resized = F.interpolate(
            coarse_features, size=low_level_features.shape[2:], mode="bilinear", align_corners=False
        )
        combined_features = torch.cat([low_level_features, coarse_features_resized], dim=1)
        return [self.final(combined_features), coarse_features]

    @torch.jit.unused
    def get_output_spec(self) -> FeatureMapsSpecification:
        return self.output_spec



# Credit: https://github.com/PkuRainBow/OCNet.pytorch/blob/master/oc_module/asp_oc_block.py
import torch
import torch.nn.functional as F
from torch import nn

from ..modules.activations import ABN

__all__ = ["ObjectContextBlock", "ASPObjectContextBlock", "PyramidObjectContextBlock"]


class _SelfAttentionBlock(nn.Module):
    """
    The basic implementation for self-attention block/non-local block
    Input:
        N X C X H X W
    Parameters:
        in_channels       : the dimension of the input feature map
        key_channels      : the dimension after the key/query transform
        value_channels    : the dimension after the value transform
        scale             : choose the scale to downsample the input feature maps (save memory cost)
    Return:
        N X C X H X W
        position-aware context features.(w/o concate or add with the input)
    """

    def __init__(self, in_channels, key_channels, value_channels, out_channels=None, scale=1, abn_block=ABN):
        super(_SelfAttentionBlock, self).__init__()
        self.scale = scale
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.key_channels = key_channels
        self.value_channels = value_channels
        if out_channels is None:
            self.out_channels = in_channels
        self.pool = nn.MaxPool2d(kernel_size=(scale, scale))
        self.f_key = nn.Sequential(
            nn.Conv2d(
                in_channels=self.in_channels, out_channels=self.key_channels, kernel_size=1, stride=1, padding=0
            ),
            abn_block(self.key_channels),
        )
        self.f_query = self.f_key
        self.f_value = nn.Conv2d(
            in_channels=self.in_channels, out_channels=self.value_channels, kernel_size=1, stride=1, padding=0
        )
        self.W = nn.Conv2d(
            in_channels=self.value_channels, out_channels=self.out_channels, kernel_size=1, stride=1, padding=0
        )
        # Eugene Khvedchenya: Original implementation initialized weight of context convolution with zeros, which does not make sense to me
        # nn.init.constant(self.W.weight, 0)
        nn.init.constant_(self.W.bias, 0)

    def forward(self, x):  # skipcq: PYL-W0221
        batch_size, h, w = x.size(0), x.size(2), x.size(3)
        if self.scale > 1:
            x = self.pool(x)

        value = self.f_value(x).view(batch_size, self.value_channels, -1)
        value = value.permute(0, 2, 1)
        query = self.f_query(x).view(batch_size, self.key_channels, -1)
        query = query.permute(0, 2, 1)
        key = self.f_key(x).view(batch_size, self.key_channels, -1)

        sim_map = torch.matmul(query, key)
        sim_map = (self.key_channels**-0.5) * sim_map
        sim_map = F.softmax(sim_map, dim=-1)

        context = torch.matmul(sim_map, value)
        context = context.permute(0, 2, 1).contiguous()
        context = context.view(batch_size, self.value_channels, *x.size()[2:])
        context = self.W(context)
        if self.scale > 1:
            context = F.interpolate(input=context, size=(h, w), mode="bilinear", align_corners=False)
        return context


class SelfAttentionBlock2D(_SelfAttentionBlock):
    def __init__(self, in_channels, key_channels, value_channels, out_channels=None, scale=1):
        super(SelfAttentionBlock2D, self).__init__(in_channels, key_channels, value_channels, out_channels, scale)


class BaseOC_Module(nn.Module):
    """
    Implementation of the BaseOC module
    Parameters:
        in_features / out_features: the channels of the input / output feature maps.
        dropout: we choose 0.05 as the default value.
        size: you can apply multiple sizes. Here we only use one size.
    Return:
        features fused with Object context information.
    """

    def __init__(self, in_channels, out_channels, key_channels, value_channels, dropout, sizes=([1]), abn_block=ABN):
        super(BaseOC_Module, self).__init__()
        self.stages = []
        self.stages = nn.ModuleList(
            [self._make_stage(in_channels, out_channels, key_channels, value_channels, size) for size in sizes]
        )
        self.conv_bn_dropout = nn.Sequential(
            nn.Conv2d(2 * in_channels, out_channels, kernel_size=1, padding=0),
            abn_block(out_channels),
            nn.Dropout2d(dropout),
        )

    def _make_stage(self, in_channels, output_channels, key_channels, value_channels, size):
        return SelfAttentionBlock2D(in_channels, key_channels, value_channels, output_channels, size)

    def forward(self, feats):
        priors = [stage(feats) for stage in self.stages]
        context = priors[0]
        for i in range(1, len(priors)):
            context += priors[i]
        output = self.conv_bn_dropout(torch.cat([context, feats], 1))
        return output


class ObjectContextBlock(nn.Module):
    """
    Output only the context features.
    Parameters:
        in_features / out_features: the channels of the input / output feature maps.
        dropout: specify the dropout ratio
        fusion: We provide two different fusion method, "concat" or "add"
        size: we find that directly learn the attention weights on even 1/8 feature maps is hard.
    Return:
        features after "concat" or "add"
    """

    def __init__(self, in_channels, out_channels, key_channels, value_channels, dropout, sizes=([1]), abn_block=ABN):
        super(ObjectContextBlock, self).__init__()
        self.stages = []
        self.stages = nn.ModuleList(
            [self._make_stage(in_channels, out_channels, key_channels, value_channels, size) for size in sizes]
        )
        self.conv_bn_dropout = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=1, padding=0, bias=False), abn_block(out_channels)
        )

    def _make_stage(self, in_channels, output_channels, key_channels, value_channels, size):
        return SelfAttentionBlock2D(in_channels, key_channels, value_channels, output_channels, size)

    def forward(self, feats):
        priors = [stage(feats) for stage in self.stages]
        context = priors[0]
        for i in range(1, len(priors)):
            context += priors[i]
        output = self.conv_bn_dropout(context)
        return output


class ASPObjectContextBlock(nn.Module):
    def __init__(self, features, out_features=256, dilations=(12, 24, 36), abn_block=ABN, dropout=0.1):
        super(ASPObjectContextBlock, self).__init__()
        self.context = nn.Sequential(
            nn.Conv2d(features, out_features, kernel_size=3, padding=1, dilation=1, bias=False),
            abn_block(out_features),
            ObjectContextBlock(
                in_channels=out_features,
                out_channels=out_features,
                key_channels=out_features // 2,
                value_channels=out_features,
                dropout=dropout,
                sizes=([2]),
            ),
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(features, out_features, kernel_size=1, padding=0, dilation=1, bias=False),
            abn_block(out_features),
        )
        self.conv3 = nn.Sequential(
            nn.Conv2d(features, out_features, kernel_size=3, padding=dilations[0], dilation=dilations[0], bias=False),
            abn_block(out_features),
        )
        self.conv4 = nn.Sequential(
            nn.Conv2d(features, out_features, kernel_size=3, padding=dilations[1], dilation=dilations[1], bias=False),
            abn_block(out_features),
        )
        self.conv5 = nn.Sequential(
            nn.Conv2d(features, out_features, kernel_size=3, padding=dilations[2], dilation=dilations[2], bias=False),
            abn_block(out_features),
        )

        self.conv_bn_dropout = nn.Sequential(
            nn.Conv2d(out_features * 5, out_features * 2, kernel_size=1, padding=0, dilation=1, bias=False),
            abn_block(out_features * 2),
            nn.Dropout2d(dropout),
        )

    def _cat_each(self, feat1, feat2, feat3, feat4, feat5):
        assert len(feat1) == len(feat2)
        z = []
        for i in range(len(feat1)):
            z.append(torch.cat((feat1[i], feat2[i], feat3[i], feat4[i], feat5[i]), dim=1))
        return z

    def forward(self, x):
        if isinstance(x, torch.Tensor):
            _, _, h, w = x.size()
        elif isinstance(x, tuple) or isinstance(x, list):
            _, _, h, w = x[0].size()
        else:
            raise RuntimeError("unknown input type")

        feat1 = self.context(x)
        feat2 = self.conv2(x)
        feat3 = self.conv3(x)
        feat4 = self.conv4(x)
        feat5 = self.conv5(x)

        if isinstance(x, torch.Tensor):
            out = torch.cat((feat1, feat2, feat3, feat4, feat5), dim=1)
        elif isinstance(x, tuple) or isinstance(x, list):
            out = self._cat_each(feat1, feat2, feat3, feat4, feat5)
        else:
            raise RuntimeError("unknown input type")

        output = self.conv_bn_dropout(out)
        return output


class _PyramidSelfAttentionBlock(nn.Module):
    """
    The basic implementation for self-attention block/non-local block
    Input:
        N X C X H X W
    Parameters:
        in_channels       : the dimension of the input feature map
        key_channels      : the dimension after the key/query transform
        value_channels    : the dimension after the value transform
        scale             : choose the scale to downsample the input feature maps
    Return:
        N X C X H X W
        position-aware context features.(w/o concate or add with the input)
    """

    def __init__(self, in_channels, key_channels, value_channels, out_channels=None, scale=1, abn_block=ABN):
        super(_PyramidSelfAttentionBlock, self).__init__()
        self.scale = scale
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.key_channels = key_channels
        self.value_channels = value_channels
        if out_channels is None:
            self.out_channels = in_channels
        self.f_key = nn.Sequential(
            nn.Conv2d(
                in_channels=self.in_channels,
                out_channels=self.key_channels,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            ),
            abn_block(self.key_channels),
        )
        self.f_query = self.f_key
        self.f_value = nn.Conv2d(
            in_channels=self.in_channels, out_channels=self.value_channels, kernel_size=1, stride=1, padding=0
        )
        self.W = nn.Conv2d(
            in_channels=self.value_channels, out_channels=self.out_channels, kernel_size=1, stride=1, padding=0
        )
        # nn.init.constant(self.W.weight, 0)
        nn.init.constant(self.W.bias, 0)

    def forward(self, x):
        batch_size, _, h, w = x.size(0), x.size(1), x.size(2), x.size(3)

        local_x = []
        local_y = []
        step_h, step_w = h // self.scale, w // self.scale
        for i in range(0, self.scale):
            for j in range(0, self.scale):
                start_x, start_y = i * step_h, j * step_w
                end_x, end_y = min(start_x + step_h, h), min(start_y + step_w, w)
                if i == (self.scale - 1):
                    end_x = h
                if j == (self.scale - 1):
                    end_y = w
                local_x += [start_x, end_x]
                local_y += [start_y, end_y]

        value = self.f_value(x)
        query = self.f_query(x)
        key = self.f_key(x)

        local_list = []
        local_block_cnt = 2 * self.scale * self.scale
        for i in range(0, local_block_cnt, 2):
            value_local = value[:, :, local_x[i] : local_x[i + 1], local_y[i] : local_y[i + 1]]
            query_local = query[:, :, local_x[i] : local_x[i + 1], local_y[i] : local_y[i + 1]]
            key_local = key[:, :, local_x[i] : local_x[i + 1], local_y[i] : local_y[i + 1]]

            h_local, w_local = value_local.size(2), value_local.size(3)
            value_local = value_local.contiguous().view(batch_size, self.value_channels, -1)
            value_local = value_local.permute(0, 2, 1)

            query_local = query_local.contiguous().view(batch_size, self.key_channels, -1)
            query_local = query_local.permute(0, 2, 1)
            key_local = key_local.contiguous().view(batch_size, self.key_channels, -1)

            sim_map = torch.matmul(query_local, key_local)
            sim_map = (self.key_channels**-0.5) * sim_map
            sim_map = F.softmax(sim_map, dim=-1)

            context_local = torch.matmul(sim_map, value_local)
            context_local = context_local.permute(0, 2, 1).contiguous()
            context_local = context_local.view(batch_size, self.value_channels, h_local, w_local)
            local_list.append(context_local)

        context_list = []
        for i in range(0, self.scale):
            row_tmp = []
            for j in range(0, self.scale):
                row_tmp.append(local_list[j + i * self.scale])
            context_list.append(torch.cat(row_tmp, 3))

        context = torch.cat(context_list, 2)
        context = self.W(context)

        return context


class PyramidSelfAttentionBlock2D(_PyramidSelfAttentionBlock):
    def __init__(self, in_channels, key_channels, value_channels, out_channels=None, scale=1):
        super(PyramidSelfAttentionBlock2D, self).__init__(
            in_channels, key_channels, value_channels, out_channels, scale
        )


class PyramidObjectContextBlock(nn.Module):
    """
    Output the combination of the context features and the original features.
    Parameters:
        in_features / out_features: the channels of the input / output feature maps.
        dropout: specify the dropout ratio
        size: we find that directly learn the attention weights on even 1/8 feature maps is hard.
    Return:
        features after "concat" or "add"
    """

    def __init__(self, in_channels, out_channels, dropout=0.05, sizes=([1, 2, 3, 6]), abn_block=ABN):
        super(PyramidObjectContextBlock, self).__init__()
        self.group = len(sizes)
        self.stages = []
        self.stages = nn.ModuleList(
            [self._make_stage(in_channels, out_channels, in_channels // 2, in_channels, size) for size in sizes]
        )
        self.conv_bn_dropout = nn.Sequential(
            nn.Conv2d(2 * in_channels * self.group, out_channels, kernel_size=1, padding=0, bias=False),
            abn_block(out_channels),
            nn.Dropout2d(dropout),
        )
        self.up_dr = nn.Sequential(
            nn.Conv2d(in_channels, in_channels * self.group, kernel_size=1, padding=0, bias=False),
            abn_block(in_channels * self.group),
        )

    def _make_stage(self, in_channels, output_channels, key_channels, value_channels, size):
        return PyramidSelfAttentionBlock2D(in_channels, key_channels, value_channels, output_channels, size)

    def forward(self, feats):
        priors = [stage(feats) for stage in self.stages]
        context = [self.up_dr(feats)]
        for i in range(0, len(priors)):
            context += [priors[i]]
        output = self.conv_bn_dropout(torch.cat(context, 1))
        return output



from typing import List, Union, Type, Tuple, Mapping

import torch
import logging

from torch import nn

from pytorch_toolbelt.modules.unet import UnetBlock, UnetResidualBlock
from pytorch_toolbelt.modules.interfaces import AbstractDecoder, FeatureMapsSpecification
from pytorch_toolbelt.modules.upsample import (
    AbstractResizeLayer,
    UpsampleLayerType,
    instantiate_upsample_block,
    DeconvolutionUpsample2d,
)
from pytorch_toolbelt.modules.normalization import NORM_BATCH, instantiate_normalization_block
from pytorch_toolbelt.modules.activations import ACT_RELU

__all__ = ["UNetDecoder"]

logger = logging.getLogger(__name__)


class UNetDecoder(AbstractDecoder):
    def __init__(
        self,
        input_spec: FeatureMapsSpecification,
        out_channels: Union[Tuple[int, ...], List[int]],
        block_type: Union[Type[UnetBlock], Type[UnetResidualBlock]] = UnetBlock,
        upsample_block: Union[UpsampleLayerType, Type[AbstractResizeLayer]] = UpsampleLayerType.BILINEAR,
        upsample_kwargs: Union[None, Mapping] = None,
        activation: str = ACT_RELU,
        normalization: str = NORM_BATCH,
        block_kwargs=None,
        unet_block=None,
        num_blocks_per_stage: Union[None, int, Tuple[int, ...]] = None,
    ):
        num_stages = len(input_spec) - 1  # Number of outputs is one less than encoder layers

        if upsample_kwargs is None:
            upsample_kwargs = {}

        if unet_block is not None:
            logger.warning("unet_block argument is deprecated, use block_type instead", DeprecationWarning)
            block_type = unet_block

        if num_blocks_per_stage is None:
            num_blocks_per_stage = 1

        if isinstance(num_blocks_per_stage, int):
            num_blocks_per_stage = (num_blocks_per_stage,) * num_stages

        num_blocks_per_stage = tuple(num_blocks_per_stage)

        if len(num_blocks_per_stage) != num_stages:
            raise ValueError(f"num_blocks_per_stage must have length of {num_stages}")

        if len(out_channels) != num_stages:
            raise ValueError(f"decoder_features must have length of {num_stages}")

        super().__init__(input_spec)
        if block_kwargs is None:
            block_kwargs = {
                "activation": activation,
                "normalization": normalization,
            }

        blocks = []
        upsamples = []

        in_channels_for_upsample_block = input_spec.channels[-1]

        for block_index in reversed(range(num_stages)):
            features_from_encoder = input_spec.channels[block_index]

            scale_factor = input_spec.strides[block_index + 1] // input_spec.strides[block_index]
            upsample_layer: AbstractResizeLayer = instantiate_upsample_block(
                upsample_block,
                in_channels=in_channels_for_upsample_block,
                scale_factor=scale_factor,
                **upsample_kwargs,
            )

            upsamples.append(upsample_layer)
            out_channels_from_upsample_block = upsample_layer.out_channels

            in_channels = features_from_encoder + out_channels_from_upsample_block

            stage = self._build_stage(
                in_channels, out_channels[block_index], block_type, block_kwargs, num_blocks_per_stage[block_index]
            )
            blocks.append(stage)

            in_channels_for_upsample_block = out_channels[block_index]

        self.blocks = nn.ModuleList(blocks)
        self.upsamples = nn.ModuleList(upsamples)
        self.output_spec = FeatureMapsSpecification(channels=out_channels, strides=input_spec.strides[:-1])

    def _build_stage(
        self, in_channels: int, out_channels: int, block_type: Type, block_kwargs: Mapping, num_blocks: int
    ):
        blocks = []
        for _ in range(num_blocks):
            blocks.append(block_type(in_channels, out_channels, **block_kwargs))
            in_channels = out_channels
        if num_blocks == 1:
            return blocks[0]
        return nn.Sequential(*blocks)

    @torch.jit.unused
    def get_output_spec(self) -> FeatureMapsSpecification:
        return self.output_spec

    def forward(self, feature_maps: List[torch.Tensor]) -> List[torch.Tensor]:
        x = feature_maps[-1]
        outputs = []
        num_feature_maps = len(feature_maps)
        for index, (upsample_block, decoder_block) in enumerate(zip(self.upsamples, self.blocks)):
            encoder_input = feature_maps[num_feature_maps - index - 2]

            x = upsample_block(x, output_size=encoder_input.size()[2:])

            x = torch.cat([x, encoder_input], dim=1)
            x = decoder_block(x)
            outputs.append(x)

        # Returns list of tensors in same order as input (fine-to-coarse)
        return outputs[::-1]



"""Wrappers for different backbones for models that follows Encoder-Decoder architecture.

Encodes listed here provides easy way to swap backbone of classification/segmentation/detection model.
"""

import math
import warnings
from typing import List, Union, Tuple, Iterable, Any

import torch
from torch import nn, Tensor

__all__ = ["EncoderModule", "_take", "_take_ints", "_take_tensors", "make_n_channel_input"]

from pytorch_toolbelt.utils.support import pytorch_toolbelt_deprecated
from pytorch_toolbelt.modules.interfaces import (
    HasOutputFeaturesSpecification,
    FeatureMapsSpecification,
    AbstractEncoder,
)


def _take(elements: List[Any], indexes: List[int]) -> List[Any]:
    selected = []
    for i in indexes:
        selected.append(elements[i])
    return selected


def _take_ints(elements: List[int], indexes: List[int]) -> List[int]:
    selected: List[int] = []
    for i in indexes:
        selected.append(elements[i])
    return selected


def _take_tensors(elements: List[Tensor], indexes: List[int]) -> List[Tensor]:
    selected: List[Tensor] = []
    for i in indexes:
        selected.append(elements[i])
    return selected


def make_n_channel_input_conv(
    conv: Union[nn.Conv1d, nn.Conv2d, nn.Conv3d], in_channels: int, mode="auto", **kwargs
) -> Union[nn.Conv1d, nn.Conv2d, nn.Conv3d]:
    """
    Create convolution block with same parameters and desired number of channels.

    Args:
        conv: Input nn.Conv2D object to copy settings/weights from
        in_channels: Desired number of input channels
        mode:
        **kwargs: Optional overrides for Conv2D parameters
    """
    conv_cls = conv.__class__

    if conv.in_channels == in_channels:
        warnings.warn("make_n_channel_input call is spurious")
        return conv

    new_conv = conv_cls(
        in_channels,
        out_channels=conv.out_channels,
        kernel_size=kwargs.get("kernel_size", conv.kernel_size),
        stride=kwargs.get("stride", conv.stride),
        padding=kwargs.get("padding", conv.padding),
        dilation=kwargs.get("dilation", conv.dilation),
        groups=kwargs.get("groups", conv.groups),
        bias=kwargs.get("bias", conv.bias is not None),
        padding_mode=kwargs.get("padding_mode", conv.padding_mode),
    )

    w = conv.weight
    if in_channels > conv.in_channels:
        n = math.ceil(in_channels / float(conv.in_channels))
        w = torch.cat([w] * n, dim=1)
        w = w[:, :in_channels, ...]
        new_conv.weight = nn.Parameter(w, requires_grad=True)
    else:
        w = w[:, 0:in_channels, ...]
        new_conv.weight = nn.Parameter(w, requires_grad=True)

    return new_conv


def make_n_channel_input(conv: nn.Module, in_channels: int, mode="auto", **kwargs) -> nn.Module:
    """
    Create convolution block with same parameters and desired number of channels.

    Args:
        conv: Input nn.Conv2D object to copy settings/weights from
        in_channels: Desired number of input channels
        mode:
        **kwargs: Optional overrides for Conv2D parameters
    """
    if isinstance(conv, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
        return make_n_channel_input_conv(conv, in_channels=in_channels, mode=mode, **kwargs)

    raise ValueError(f"Unsupported class {conv.__class__.__name__}")


class EncoderModule(AbstractEncoder):
    __constants__ = ["_layers", "_output_strides", "_output_filters"]

    def __init__(self, channels: List[int], strides: List[int], layers: List[int]):
        super().__init__()
        if len(channels) != len(strides):
            raise ValueError("Number of channels must be equal to number of strides")

        self._layers = list(layers)
        self.output_spec = FeatureMapsSpecification(
            channels=_take_ints(channels, layers), strides=_take_ints(strides, layers)
        )

    def forward(self, x: Tensor) -> List[Tensor]:  # skipcq: PYL-W0221
        output_features = []
        for layer in self.encoder_layers:
            output = layer(x)
            output_features.append(output)
            x = output
        # Return only features that were requested
        return _take_tensors(output_features, self._layers)

    @property
    @torch.jit.unused
    def channels(self) -> Tuple[int, ...]:
        return self.get_output_spec().channels

    @property
    @torch.jit.unused
    def strides(self) -> Tuple[int, ...]:
        return self.get_output_spec().strides

    @torch.jit.unused
    def set_trainable(self, trainable):
        for param in self.parameters():
            param.requires_grad = bool(trainable)

    @torch.jit.unused
    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        """
        Change number of channels expected in the input tensor. By default,
        all encoders assume 3-channel image in BCHW notation with C=3.
        This method changes first convolution to have user-defined number of
        channels as input.
        """
        raise NotImplementedError

    @torch.jit.unused
    def get_output_spec(self) -> FeatureMapsSpecification:
        return self.output_spec



from .common import EncoderModule, _take, make_n_channel_input
from ..backbone.mobilenet import MobileNetV2

__all__ = ["MobilenetV2Encoder", "MobileNetV3Large", "MobileNetV3Small"]


class MobilenetV2Encoder(EncoderModule):
    def __init__(self, layers=[2, 3, 5, 7], activation="relu6"):
        super().__init__([32, 16, 24, 32, 64, 96, 160, 320], [2, 2, 4, 8, 16, 16, 32, 32], layers)
        encoder = MobileNetV2(activation=activation)

        self.layer0 = encoder.layer0
        self.layer1 = encoder.layer1
        self.layer2 = encoder.layer2
        self.layer3 = encoder.layer3
        self.layer4 = encoder.layer4
        self.layer5 = encoder.layer5
        self.layer6 = encoder.layer6
        self.layer7 = encoder.layer7

    @property
    def encoder_layers(self):
        return [self.layer0, self.layer1, self.layer2, self.layer3, self.layer4, self.layer5, self.layer6, self.layer7]

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.layer0.conv1 = make_n_channel_input(self.layer0.conv1, input_channels, mode)
        return self


class MobileNetV3Large(EncoderModule):
    def __init__(self, pretrained=True, layers=None):
        from torchvision.models.mobilenetv3 import mobilenet_v3_large

        encoder = mobilenet_v3_large(pretrained=pretrained)

        if layers is None:
            layers = [1, 2, 3, 4]
        super().__init__([16, 24, 40, 112, 960], [2, 4, 8, 16, 32], layers)
        from torchvision.models import MobileNetV3

        self.layer0 = encoder.features[0:2]
        self.layer1 = encoder.features[2:4]
        self.layer2 = encoder.features[4:7]
        self.layer3 = encoder.features[7:13]
        self.layer4 = encoder.features[13:]

    @property
    def encoder_layers(self):
        return [self.layer0, self.layer1, self.layer2, self.layer3, self.layer4]

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.layer0[0][0] = make_n_channel_input(self.layer0[0][0], input_channels, mode=mode, **kwargs)
        return self


class MobileNetV3Small(EncoderModule):
    def __init__(self, pretrained=True, layers=None):
        from torchvision.models.mobilenetv3 import mobilenet_v3_small

        encoder = mobilenet_v3_small(pretrained=pretrained)

        if layers is None:
            layers = [1, 2, 3, 4]
        super().__init__([16, 16, 24, 48, 576], [2, 4, 8, 16, 32], layers)
        from torchvision.models import MobileNetV3

        self.layer0 = encoder.features[0:1]
        self.layer1 = encoder.features[1:2]
        self.layer2 = encoder.features[2:4]
        self.layer3 = encoder.features[4:9]
        self.layer4 = encoder.features[9:]

    @property
    def encoder_layers(self):
        return [self.layer0, self.layer1, self.layer2, self.layer3, self.layer4]

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.layer0[0][0] = make_n_channel_input(self.layer0[0][0], input_channels, mode=mode, **kwargs)
        return self



from collections import defaultdict

import cv2
import numpy as np
import pytest
import torch
from torch import nn

from pytorch_toolbelt.inference import tta
from pytorch_toolbelt.utils.torch_utils import to_numpy

skip_if_no_cuda = pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available")


class NoOp(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        return input


class SumAll(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        return input.sum(dim=[1, 2, 3])


def test_d4_image2mask():
    x = torch.rand((4, 3, 224, 224))
    model = NoOp()

    output = tta.d4_image2mask(model, x)
    np.testing.assert_allclose(to_numpy(output), to_numpy(x), atol=1e-6, rtol=1e-6)


def test_flips_image2mask_v2():
    x = torch.rand((4, 3, 224, 224))
    x_a = tta.flips_image_augment(x)
    y = tta.flips_image_deaugment(x_a)

    np.testing.assert_allclose(to_numpy(y), to_numpy(x), atol=1e-6, rtol=1e-6)


def test_d2_image2mask_v2():
    x = torch.rand((4, 3, 224, 224))
    x_a = tta.d2_image_augment(x)
    y = tta.d2_image_deaugment(x_a)

    np.testing.assert_allclose(to_numpy(y), to_numpy(x), atol=1e-6, rtol=1e-6)


def test_d4_image2mask_v2():
    x = torch.rand((4, 3, 224, 224))
    x_a = tta.d4_image_augment(x)
    y = tta.d4_image_deaugment(x_a)

    np.testing.assert_allclose(to_numpy(y), to_numpy(x), atol=1e-6, rtol=1e-6)


def test_fliplr_image2mask():
    x = torch.rand((4, 3, 224, 224))
    model = NoOp()

    output = tta.fliplr_image2mask(model, x)
    np.testing.assert_allclose(to_numpy(output), to_numpy(x), atol=1e-6, rtol=1e-6)


def test_d4_image2label():
    x = torch.tensor([[1, 2, 3, 4], [5, 6, 7, 8], [9, 0, 1, 2], [3, 4, 5, 6]]).unsqueeze(0).unsqueeze(0).float()
    model = SumAll()

    output = tta.d4_image2label(model, x)
    expected = int(x.sum())

    assert int(output) == expected


def test_fliplr_image2label():
    x = torch.tensor([[1, 2, 3, 4], [5, 6, 7, 8], [9, 0, 1, 2], [3, 4, 5, 6]]).unsqueeze(0).unsqueeze(0).float()
    model = SumAll()

    output = tta.fliplr_image2label(model, x)
    expected = int(x.sum())

    assert int(output) == expected


def test_fivecrop_image2label():
    x = torch.tensor([[1, 2, 3, 4], [5, 6, 7, 8], [9, 0, 1, 2], [3, 4, 5, 6]]).unsqueeze(0).unsqueeze(0).float()
    model = SumAll()

    output = tta.fivecrop_image2label(model, x, (2, 2))
    expected = ((1 + 2 + 5 + 6) + (3 + 4 + 7 + 8) + (9 + 0 + 3 + 4) + (1 + 2 + 5 + 6) + (6 + 7 + 0 + 1)) / 5

    assert int(output) == expected


def test_tencrop_image2label():
    x = torch.tensor([[1, 2, 3, 4], [5, 6, 7, 8], [9, 0, 1, 2], [3, 4, 5, 6]]).unsqueeze(0).unsqueeze(0).float()
    model = SumAll()

    output = tta.tencrop_image2label(model, x, (2, 2))
    expected = (2 * ((1 + 2 + 5 + 6) + (3 + 4 + 7 + 8) + (9 + 0 + 3 + 4) + (1 + 2 + 5 + 6) + (6 + 7 + 0 + 1))) / 10

    assert int(output) == expected



# from typing import List
#
# import torch
# import torch.nn.functional as F
# from torch import nn
#
#
# class PPMDecoder(AbstractDecoder):
#     """
#     Pyramid pooling decoder module
#
#     https://github.com/CSAILVision/semantic-segmentation-pytorch/blob/42b7567a43b1dab568e2bbfcbc8872778fbda92a/models/models.py
#     """
#
#     def __init__(self, feature_maps: List[int], num_classes=150, channels=512, pool_scales=(1, 2, 3, 6)):
#         super(PPMDecoder, self).__init__()
#
#         fc_dim = feature_maps[-1]
#         self.ppm = []
#         for scale in pool_scales:
#             self.ppm.append(
#                 nn.Sequential(
#                     nn.AdaptiveAvgPool2d(scale),
#                     nn.Conv2d(fc_dim, channels, kernel_size=1, bias=False),
#                     nn.BatchNorm2d(channels),
#                     nn.ReLU(inplace=True),
#                 )
#             )
#         self.ppm = nn.ModuleList(self.ppm)
#
#         self.conv_last = nn.Sequential(
#             nn.Conv2d(fc_dim + len(pool_scales) * channels, channels, kernel_size=3, padding=1, bias=False),
#             nn.BatchNorm2d(channels),
#             nn.ReLU(inplace=True),
#             nn.Dropout2d(0.1),
#             nn.Conv2d(channels, num_classes, kernel_size=1),
#         )
#
#     def forward(self, feature_maps: List[torch.Tensor]):
#         last_fm = feature_maps[-1]
#
#         input_size = last_fm.size()
#         ppm_out = [last_fm]
#         for pool_scale in self.ppm:
#             input_pooled = pool_scale(last_fm)
#             input_pooled = F.interpolate(input_pooled, size=input_size[2:], mode="bilinear", align_corners=False)
#             ppm_out.append(input_pooled)
#         ppm_out = torch.cat(ppm_out, dim=1)
#
#         x = self.conv_last(ppm_out)
#         return x



from typing import Union

import torch.nn.init
from torch import nn

__all__ = ["first_class_background_init"]


def first_class_background_init(
    module: Union[nn.Linear, nn.Conv1d, nn.Conv2d, nn.Conv3d], background_prob: float = 0.95
):
    """
    Initialize weights of the Linear or Conv layer with zeros and bias with values of
    [logit(bg_prob), logit(fg_prob), logit(fg_prob),  ...]
    """
    bg_bias = torch.logit(torch.tensor(background_prob))
    fg_bias = torch.logit(torch.tensor((1 - background_prob)))

    torch.nn.init.zeros_(module.weight)
    torch.nn.init.constant_(module.bias, bg_bias)
    torch.nn.init.constant_(module.bias[1:], fg_bias)



from collections import OrderedDict
from typing import List

import torch
from torch import nn
from torchvision.models import densenet121, densenet161, densenet169, densenet201, DenseNet

from .common import EncoderModule, _take, make_n_channel_input

__all__ = ["DenseNetEncoder", "DenseNet121Encoder", "DenseNet169Encoder", "DenseNet161Encoder", "DenseNet201Encoder"]


class DenseNetEncoder(EncoderModule):
    def __init__(
        self, densenet: DenseNet, strides: List[int], channels: List[int], layers: List[int], first_avg_pool=False
    ):
        if layers is None:
            layers = [1, 2, 3, 4]

        super().__init__(channels, strides, layers)

        def except_pool(block: nn.Module):
            del block.pool
            return block

        self.layer0 = nn.Sequential(
            OrderedDict(
                [
                    ("conv0", densenet.features.conv0),
                    ("bn0", densenet.features.norm0),
                    ("act0", densenet.features.relu0),
                ]
            )
        )

        self.avg_pool = nn.AvgPool2d(kernel_size=2, stride=2)
        self.pool0 = self.avg_pool if first_avg_pool else densenet.features.pool0

        self.layer1 = nn.Sequential(densenet.features.denseblock1, except_pool(densenet.features.transition1))

        self.layer2 = nn.Sequential(densenet.features.denseblock2, except_pool(densenet.features.transition2))

        self.layer3 = nn.Sequential(densenet.features.denseblock3, except_pool(densenet.features.transition3))

        self.layer4 = nn.Sequential(densenet.features.denseblock4)

        self._output_strides = _take(strides, layers)
        self._output_filters = _take(channels, layers)

    @property
    def encoder_layers(self):
        return [self.layer0, self.layer1, self.layer2, self.layer3, self.layer4]

    @property
    @torch.jit.unused
    def strides(self):
        return self._output_strides

    @property
    @torch.jit.unused
    def channels(self):
        return self._output_filters

    def forward(self, x):
        output_features = []
        for layer in self.encoder_layers:
            output = layer(x)
            output_features.append(output)

            if layer == self.layer0:
                # Fist maxpool operator is not a part of layer0 because we want that layer0 output to have stride of 2
                output = self.pool0(output)
            else:
                output = self.avg_pool(output)

            x = output

        # Return only features that were requested
        return _take(output_features, self._layers)

    @torch.jit.unused
    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.layer0.conv0 = make_n_channel_input(self.layer0.conv0, input_channels, mode=mode, **kwargs)
        return self


class DenseNet121Encoder(DenseNetEncoder):
    def __init__(self, layers=None, pretrained=True, memory_efficient=False, first_avg_pool=False):
        densenet = densenet121(pretrained=pretrained, memory_efficient=memory_efficient)
        strides = [2, 4, 8, 16, 32]
        channels = [64, 128, 256, 512, 1024]
        super().__init__(densenet, strides, channels, layers, first_avg_pool)


class DenseNet161Encoder(DenseNetEncoder):
    def __init__(self, layers=None, pretrained=True, memory_efficient=False, first_avg_pool=False):
        densenet = densenet161(pretrained=pretrained, memory_efficient=memory_efficient)
        strides = [2, 4, 8, 16, 32]
        channels = [96, 192, 384, 1056, 2208]
        super().__init__(densenet, strides, channels, layers, first_avg_pool)


class DenseNet169Encoder(DenseNetEncoder):
    def __init__(self, layers=None, pretrained=True, memory_efficient=False, first_avg_pool=False):
        densenet = densenet169(pretrained=pretrained, memory_efficient=memory_efficient)
        strides = [2, 4, 8, 16, 32]
        channels = [64, 128, 256, 640, 1664]
        super().__init__(densenet, strides, channels, layers, first_avg_pool)


class DenseNet201Encoder(DenseNetEncoder):
    def __init__(self, layers=None, pretrained=True, memory_efficient=False, first_avg_pool=False):
        densenet = densenet201(pretrained=pretrained, memory_efficient=memory_efficient)
        strides = [2, 4, 8, 16, 32]
        channels = [64, 128, 256, 896, 1920]
        super().__init__(densenet, strides, channels, layers, first_avg_pool)



import pytest
import torch

import pytorch_toolbelt.modules.encoders as E
from pytorch_toolbelt.modules import AbstractEncoder
from pytorch_toolbelt.modules.backbone.inceptionv4 import inceptionv4
from pytorch_toolbelt.utils.torch_utils import maybe_cuda, count_parameters, describe_outputs
from pytorch_toolbelt.modules.encoders import timm

skip_if_no_cuda = pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available")


@pytest.mark.parametrize(
    ["encoder", "encoder_params"],
    [[E.MobileNetV3Large, {"layers": [0, 1, 2, 3, 4]}], [E.MobileNetV3Small, {"layers": [0, 1, 2, 3, 4]}]],
)
@torch.no_grad()
@skip_if_no_cuda
def test_mobilenetv3_encoders(encoder: E.EncoderModule, encoder_params):
    net = encoder(**encoder_params).eval().change_input_channels(1)
    print(net.__class__.__name__, count_parameters(net))
    print(net.strides)
    print(net.channels)
    x = torch.rand((4, 1, 384, 256))
    x = maybe_cuda(x)
    net = maybe_cuda(net)
    output = net(x)
    assert len(output) == len(net.channels)
    for feature_map, expected_stride, expected_channels in zip(output, net.strides, net.channels):
        assert feature_map.size(1) == expected_channels
        assert feature_map.size(2) * expected_stride == 384
        assert feature_map.size(3) * expected_stride == 256


@pytest.mark.parametrize(
    ["encoder", "encoder_params"],
    [
        [E.MobilenetV2Encoder, {}],
        [E.Resnet34Encoder, {"pretrained": False}],
        [E.Resnet50Encoder, {"pretrained": False}],
        [E.SEResNeXt50Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [E.SEResnet50Encoder, {"pretrained": False}],
        [E.Resnet152Encoder, {"pretrained": False}],
        [E.Resnet101Encoder, {"pretrained": False}],
        [E.SEResnet152Encoder, {"pretrained": False}],
        [E.SEResNeXt101Encoder, {"pretrained": False}],
        [E.SEResnet101Encoder, {"pretrained": False}],
        [E.SENet154Encoder, {"pretrained": False}],
        [E.WiderResnet16Encoder, {}],
        [E.WiderResnet20Encoder, {}],
        [E.WiderResnet38Encoder, {}],
        [E.WiderResnet16A2Encoder, {}],
        [E.WiderResnet20A2Encoder, {}],
        [E.WiderResnet38A2Encoder, {}],
        [E.DenseNet121Encoder, {"pretrained": False}],
        [E.DenseNet161Encoder, {"pretrained": False}],
        [E.DenseNet169Encoder, {"pretrained": False}],
        [E.DenseNet201Encoder, {"pretrained": False}],
    ],
)
@torch.no_grad()
@skip_if_no_cuda
def test_encoders(encoder: E.EncoderModule, encoder_params):
    net = encoder(**encoder_params).eval()
    print(net.__class__.__name__, count_parameters(net))
    print(net.strides)
    print(net.channels)
    x = torch.rand((4, 3, 256, 256))
    x = maybe_cuda(x)
    net = maybe_cuda(net)
    output = net(x)
    assert len(output) == len(net.channels)
    for feature_map, expected_stride, expected_channels in zip(output, net.strides, net.channels):
        assert feature_map.size(1) == expected_channels
        assert feature_map.size(2) * expected_stride == 256
        assert feature_map.size(3) * expected_stride == 256


@torch.no_grad()
@skip_if_no_cuda
def test_unet_encoder():
    net = E.UnetEncoder().eval()
    print(net.__class__.__name__, count_parameters(net))
    print(net.strides)
    print(net.channels)
    x = torch.rand((4, 3, 256, 256))
    x = maybe_cuda(x)
    net = maybe_cuda(net)
    output = net(x)
    assert len(output) == len(net.channels)
    for feature_map, expected_stride, expected_channels in zip(output, net.strides, net.channels):
        print(feature_map.size(), feature_map.mean(), feature_map.std())
        assert feature_map.size(1) == expected_channels
        assert feature_map.size(2) * expected_stride == 256
        assert feature_map.size(3) * expected_stride == 256


@torch.no_grad()
@skip_if_no_cuda
def test_inceptionv4_encoder():
    backbone = inceptionv4(pretrained=False)
    backbone.last_linear = None

    net = E.InceptionV4Encoder(pretrained=False, layers=[0, 1, 2, 3, 4]).cuda()

    print(count_parameters(net))

    x = torch.randn((4, 3, 512, 512)).cuda()

    out = net(x)
    for fm in out:
        print(fm.size())


@torch.no_grad()
@skip_if_no_cuda
def test_densenet():
    from torchvision.models import densenet121

    net1 = E.DenseNet121Encoder(pretrained=False)
    net2 = densenet121(pretrained=False)
    net2.classifier = None

    print(count_parameters(net1), count_parameters(net2))


@pytest.mark.parametrize(
    ["encoder", "encoder_params"],
    [
        [E.HRNetV2Encoder18, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [E.HRNetV2Encoder34, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [E.HRNetV2Encoder48, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
    ],
)
@torch.no_grad()
@skip_if_no_cuda
def test_hrnet_encoder(encoder: E.EncoderModule, encoder_params):
    net = encoder(**encoder_params).eval()
    print(net.__class__.__name__, count_parameters(net))
    print(net.strides)
    print(net.channels)
    x = torch.rand((4, 3, 256, 256))
    x = maybe_cuda(x)
    net = maybe_cuda(net)
    output = net(x)
    assert len(output) == len(net.channels)
    for feature_map, expected_stride, expected_channels in zip(output, net.strides, net.channels):
        assert feature_map.size(1) == expected_channels
        assert feature_map.size(2) * expected_stride == 256
        assert feature_map.size(3) * expected_stride == 256


@pytest.mark.parametrize(
    ["encoder", "encoder_params"],
    [
        [E.XResNet18Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [E.XResNet34Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [E.XResNet50Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [E.XResNet101Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [E.XResNet152Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [E.SEXResNet18Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [E.SEXResNet34Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [E.SEXResNet50Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [E.SEXResNet101Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [E.SEXResNet152Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
    ],
)
@torch.no_grad()
@skip_if_no_cuda
def test_xresnet_encoder(encoder, encoder_params):
    net = encoder(**encoder_params).eval()
    print(net.__class__.__name__, count_parameters(net))
    print(net.strides)
    print(net.channels)
    x = torch.rand((4, 3, 256, 256))
    x = maybe_cuda(x)
    net = maybe_cuda(net)
    output = net(x)
    assert len(output) == len(net.channels)
    for feature_map, expected_stride, expected_channels in zip(output, net.strides, net.channels):
        assert feature_map.size(1) == expected_channels
        assert feature_map.size(2) * expected_stride == 256
        assert feature_map.size(3) * expected_stride == 256


@pytest.mark.parametrize(
    ["encoder", "encoder_params"],
    [
        [timm.B0Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.B0Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4], "first_conv_stride_one": True}],
        [timm.B1Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.B1Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4], "first_conv_stride_one": True}],
        [timm.B2Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.B3Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.B4Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.B5Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.B6Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.B7Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.MixNetXLEncoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.SKResNet18Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.SKResNeXt50Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.SWSLResNeXt101Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.TimmResnet200D, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.HRNetW18Encoder, {"pretrained": False}],
        [timm.HRNetW32Encoder, {"pretrained": False}],
        [timm.HRNetW48Encoder, {"pretrained": False}],
        [timm.DPN68Encoder, {"pretrained": False}],
        [timm.DPN68BEncoder, {"pretrained": False}],
        [timm.DPN92Encoder, {"pretrained": False}],
        [timm.DPN107Encoder, {"pretrained": False}],
        [timm.DPN131Encoder, {"pretrained": False}],
        [timm.NFNetF0Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.NFNetF1Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.NFNetF2Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.NFNetF3Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.NFNetF4Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.NFNetF5Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.NFNetF6Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.NFNetF7Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        #
        [timm.NFRegNetB0Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.NFRegNetB1Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.NFRegNetB2Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.NFRegNetB3Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.NFRegNetB4Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.NFRegNetB5Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        # Res2Net
        [timm.TimmRes2Net101Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.TimmRes2Next50Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        # EfficientNet V2
        [timm.TimmEfficientNetV2, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
    ],
)
@torch.no_grad()
@skip_if_no_cuda
def test_timm_encoders(encoder, encoder_params):
    net = encoder(**encoder_params).eval().change_input_channels(5)

    print(net.__class__.__name__, count_parameters(net))
    print(net.strides)
    print(net.channels)
    x = torch.rand((4, 5, 256, 256))
    x = maybe_cuda(x)
    net = maybe_cuda(net)
    output = net(x)
    assert len(output) == len(net.channels)
    for feature_map, expected_stride, expected_channels in zip(output, net.strides, net.channels):
        assert feature_map.size(0) == x.size(0)
        assert feature_map.size(1) == expected_channels
        assert feature_map.size(2) * expected_stride == x.size(2)
        assert feature_map.size(3) * expected_stride == x.size(3)

    print(describe_outputs(output))


@pytest.mark.parametrize(
    ["encoder", "encoder_params"],
    [
        [E.StackedHGEncoder, {"repeats": 1}],
        [E.StackedHGEncoder, {"repeats": 3}],
        [E.StackedHGEncoder, {"repeats": 1, "stack_level": 4}],
        [E.StackedHGEncoder, {"repeats": 3, "stack_level": 4}],
        [E.StackedHGEncoder, {"repeats": 1, "stack_level": 4, "features": 128}],
        [E.StackedHGEncoder, {"repeats": 3, "stack_level": 4, "features": 128}],
    ],
)
@torch.no_grad()
@skip_if_no_cuda
def test_hourglass_encoder(encoder, encoder_params):
    net = encoder(**encoder_params).eval()
    print(repr(net), count_parameters(net))
    print("Strides ", net.strides)
    print("Channels", net.channels)
    x = torch.rand((4, 3, 256, 256))
    x = maybe_cuda(x)
    net = maybe_cuda(net)
    output = net(x)
    assert len(output) == len(net.channels)
    for feature_map, expected_stride, expected_channels in zip(output, net.strides, net.channels):
        assert feature_map.size(1) == expected_channels
        assert feature_map.size(2) * expected_stride == 256
        assert feature_map.size(3) * expected_stride == 256


@pytest.mark.parametrize(["encoder", "encoder_params"], [[E.StackedSupervisedHGEncoder, {"supervision_channels": 1}]])
@torch.no_grad()
@skip_if_no_cuda
def test_supervised_hourglass_encoder(encoder, encoder_params):
    net = encoder(**encoder_params).eval()
    print(net.__class__.__name__, count_parameters(net))
    print(net.get_output_spec())

    x = torch.rand((4, 3, 256, 256))
    x = maybe_cuda(x)
    net = maybe_cuda(net)
    output, supervision = net(x)
    assert len(output) == len(net.channels)
    assert len(supervision) == len(net.channels) - 2

    for feature_map, expected_stride, expected_channels in zip(output, net.strides, net.channels):
        assert feature_map.size(1) == expected_channels
        assert feature_map.size(2) * expected_stride == 256
        assert feature_map.size(3) * expected_stride == 256


@pytest.mark.parametrize(
    ["encoder", "encoder_params"],
    [
        [E.SwinT, {}],
        [E.SwinS, {}],
        [E.SwinB, {}],
        [E.SwinL, {}],
    ],
)
@torch.no_grad()
@skip_if_no_cuda
def test_swin_encoder(encoder, encoder_params):
    net = encoder(**encoder_params).change_input_channels(5).eval()
    output_spec = net.get_output_spec()

    print(net.__class__.__name__, count_parameters(net))
    print(output_spec)

    x = torch.rand((4, 5, 256, 256))
    x = maybe_cuda(x)
    net = maybe_cuda(net)
    output = net(x)

    assert len(output) == len(output_spec.channels)

    for feature_map, expected_stride, expected_channels in zip(output, output_spec.strides, output_spec.channels):
        assert feature_map.size(1) == expected_channels
        assert feature_map.size(2) * expected_stride == x.size(2)
        assert feature_map.size(3) * expected_stride == x.size(3)


@pytest.mark.parametrize(
    ["encoder", "encoder_params"],
    [
        [E.MitB0Encoder, {"pretrained": True}],
        [E.MitB1Encoder, {"pretrained": True}],
        [E.MitB2Encoder, {"pretrained": True}],
        [E.MitB3Encoder, {"pretrained": True}],
        [E.MitB4Encoder, {"pretrained": True}],
        [E.MitB5Encoder, {"pretrained": True}],
    ],
)
@torch.no_grad()
@skip_if_no_cuda
def test_mit_encoder(encoder, encoder_params):
    encoder: AbstractEncoder = encoder(**encoder_params).change_input_channels(5).eval()
    print(encoder.__class__.__name__, count_parameters(encoder, human_friendly=True))

    output_spec = encoder.get_output_spec()
    print(encoder.get_output_spec())
    x = torch.rand((4, 5, 256, 256))
    x = maybe_cuda(x)
    encoder = maybe_cuda(encoder)
    output = encoder(x)

    assert len(output) == len(output_spec.channels)

    for feature_map, expected_stride, expected_channels in zip(output, output_spec.strides, output_spec.channels):
        assert feature_map.size(1) == expected_channels
        assert feature_map.size(2) * expected_stride == x.size(2)
        assert feature_map.size(3) * expected_stride == x.size(3)



from torch import nn

from pytorch_toolbelt.modules.drop_path import DropPath
from pytorch_toolbelt.modules.activations import ACT_RELU, instantiate_activation_block
from pytorch_toolbelt.modules.normalization import NORM_BATCH, instantiate_normalization_block

__all__ = ["UnetBlock", "UnetResidualBlock"]


class UnetBlock(nn.Module):
    """
    Vanilla U-Net block containing of two convolutions interleaved with batch-norm and RELU
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        activation=ACT_RELU,
        normalization=NORM_BATCH,
        normalization_kwargs=None,
        activation_kwargs=None,
    ):
        super().__init__()

        if activation_kwargs is None:
            activation_kwargs = {"inplace": True}
        if normalization_kwargs is None:
            normalization_kwargs = {}

        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1, stride=1, bias=False)
        self.norm1 = instantiate_normalization_block(normalization, out_channels, **normalization_kwargs)
        self.act1 = instantiate_activation_block(activation, **activation_kwargs)

        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1, stride=1, bias=False)
        self.norm2 = instantiate_normalization_block(normalization, out_channels, **normalization_kwargs)
        self.act2 = instantiate_activation_block(activation, **activation_kwargs)

    def forward(self, x):
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.act1(x)

        x = self.conv2(x)
        x = self.norm2(x)
        x = self.act2(x)
        return x


class UnetResidualBlock(nn.Module):
    """ """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        activation: str = ACT_RELU,
        normalization: str = NORM_BATCH,
        normalization_kwargs=None,
        activation_kwargs=None,
        drop_path_rate: float = 0.0,
    ):
        super().__init__()

        if activation_kwargs is None:
            activation_kwargs = {"inplace": True}
        if normalization_kwargs is None:
            normalization_kwargs = {}

        self.residual = (
            nn.Conv2d(in_channels, out_channels, kernel_size=1, padding=0, stride=1, bias=False)
            if in_channels != out_channels
            else nn.Identity()
        )

        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1, stride=1, bias=False)
        self.norm1 = instantiate_normalization_block(normalization, out_channels, **normalization_kwargs)
        self.act1 = instantiate_activation_block(activation, **activation_kwargs)

        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1, stride=1, bias=False)
        self.norm2 = instantiate_normalization_block(normalization, out_channels, **normalization_kwargs)
        self.act2 = instantiate_activation_block(activation, **activation_kwargs)

        self.drop_path = DropPath(drop_path_rate) if drop_path_rate > 0.0 else nn.Identity()

    def forward(self, x):
        residual = self.residual(x)

        x = self.conv1(x)
        x = self.norm1(x)
        x = self.act1(x)

        x = self.conv2(x)
        x = self.norm2(x)
        x = self.act2(self.drop_path(x) + residual)

        return x



import numpy as np

from pytorch_toolbelt.utils import plot_confusion_matrix, plot_heatmap


def test_plot_confusion_matrix():
    cm = np.random.randint(0, 7, (7, 7))

    plot_confusion_matrix(
        cm, class_names=["A", "B", "C", "D", "E", "F", "G"], normalize=True, fname="test_plot_confusion_matrix.png"
    )


def test_plot_heatmap():
    cm = np.random.randn(20, 30)

    plot_heatmap(cm, title="Test", x_label="30", y_label="20", fname="test_plot_heatmap.png", noshow=False)



from typing import Optional, Callable, Mapping

import numpy as np
from torch import nn

from pytorch_toolbelt.modules import instantiate_activation_block
from pytorch_toolbelt.modules.interfaces import AbstractHead, FeatureMapsSpecification
from pytorch_toolbelt.modules.normalization import NORM_BATCH, instantiate_normalization_block

__all__ = ["ProgressiveShuffleHead"]


class ProgressiveShuffleBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        activation: str,
        normalization: str = NORM_BATCH,
        normalization_kwargs: Optional[Mapping] = None,
    ):
        super().__init__()
        if normalization_kwargs is None:
            normalization_kwargs = {}
        self.conv = nn.Sequential(
            nn.Conv2d(in_channels, in_channels, kernel_size=3, padding=1, bias=False),
            instantiate_normalization_block(normalization, in_channels, **normalization_kwargs),
            instantiate_activation_block(activation, inplace=True),
            nn.Conv2d(in_channels, out_channels * 4, kernel_size=1, padding=0, bias=False),
        )

        self.shuffle = nn.PixelShuffle(2)

    def forward(self, x):
        return self.shuffle(self.conv(x))


def divisible(channels: int, divisor: int) -> int:
    return int(np.ceil(channels / float(divisor))) * divisor


class ProgressiveShuffleHead(AbstractHead):
    """
    Progressive Shuffle head takes the largest feature map and progressively upsample it to the original image size.
    Upsampling is done by doing [Conv3x3 + BN + Activation + Conv1x1 + PixelShuffle].
    At each stage number of channels is reduced by a factor given by `reduction_factor`.

    """

    def __init__(
        self,
        input_spec: FeatureMapsSpecification,
        num_classes: int,
        activation: str,
        dropout_rate,
        output_name: Optional[str],
        reduction_factor: int = 2,
        normalization: str = NORM_BATCH,
        normalization_kwargs: Optional[Mapping] = None,
    ):
        super().__init__(input_spec)

        self.num_classes = num_classes
        self.feature_map_index = input_spec.get_index_of_largest_feature_map()
        self.output_name = output_name

        min_stride = input_spec.strides[self.feature_map_index]
        num_upsample_blocks = int(np.log2(min_stride))
        in_channels = input_spec.channels[self.feature_map_index]
        blocks = []

        for _ in range(num_upsample_blocks):
            out_channels = divisible(in_channels / reduction_factor, 8)
            blocks.append(
                ProgressiveShuffleBlock(
                    in_channels,
                    out_channels,
                    activation=activation,
                    normalization=normalization,
                    normalization_kwargs=normalization_kwargs,
                )
            )
            in_channels = out_channels

        blocks += [
            nn.Dropout2d(p=dropout_rate),
        ]

        self.blocks = nn.Sequential(*blocks)
        self.final = nn.Conv2d(in_channels, self.num_classes, kernel_size=3, padding=1, bias=True)

        self.output_spec = FeatureMapsSpecification(channels=(num_classes,), strides=(1,))

    def forward(self, feature_maps, output_size=None):
        x = self.blocks(feature_maps[self.feature_map_index])
        output = self.final(x)

        if self.output_name is not None:
            return {self.output_name: output}
        else:
            return output

    def apply_to_final_layer(self, func: Callable[[nn.Module], None]):
        func(self.final)



import numpy as np
import pytest
import pytorch_toolbelt.losses.functional as F
import torch
from pytorch_toolbelt import losses as L
from torch import nn

from pytorch_toolbelt.losses import BinaryFocalLoss, CrossEntropyFocalLoss


def test_sigmoid_focal_loss_with_logits():
    input_good = torch.tensor([10, -10, 10]).float()
    input_bad = torch.tensor([-1, 2, 0]).float()
    target = torch.tensor([1, 0, 1])

    loss_good = F.focal_loss_with_logits(input_good, target)
    loss_bad = F.focal_loss_with_logits(input_bad, target)
    assert loss_good < loss_bad

    loss_cls = BinaryFocalLoss()
    assert loss_cls(input_good, target) < loss_cls(input_bad, target)


def test_softmax_focal_loss_with_logits():
    input_good = torch.tensor([[0, 10, 0], [10, 0, 0], [0, 0, 10]]).float()
    input_bad = torch.tensor([[0, -10, 0], [0, 10, 0], [0, 0, 10]]).float()
    target = torch.tensor([1, 0, 2]).long()

    loss_good = F.softmax_focal_loss_with_logits(input_good, target)
    loss_bad = F.softmax_focal_loss_with_logits(input_bad, target)
    assert loss_good < loss_bad

    loss_cls = CrossEntropyFocalLoss()
    assert loss_cls(input_good, target) < loss_cls(input_bad, target)


@pytest.mark.parametrize(
    ["y_true", "y_pred", "expected", "eps"],
    [
        [[1, 1, 1, 1], [1, 1, 1, 1], 1.0, 1e-5],
        [[0, 1, 1, 0], [0, 1, 1, 0], 1.0, 1e-5],
        [[1, 1, 1, 1], [1, 1, 0, 0], 0.5, 1e-5],
    ],
)
def test_soft_jaccard_score(y_true, y_pred, expected, eps):
    y_true = torch.tensor(y_true, dtype=torch.float32)
    y_pred = torch.tensor(y_pred, dtype=torch.float32)
    actual = F.soft_jaccard_score(y_pred, y_true, eps=eps)
    assert float(actual) == pytest.approx(expected, eps)


@pytest.mark.parametrize(
    ["y_true", "y_pred", "expected", "eps"],
    [
        [[[1, 1, 0, 0], [0, 0, 1, 1]], [[1, 1, 0, 0], [0, 0, 1, 1]], 1.0, 1e-5],
        [[[1, 1, 0, 0], [0, 0, 1, 1]], [[0, 0, 1, 0], [0, 1, 0, 0]], 0.0, 1e-5],
        [[[1, 1, 0, 0], [0, 0, 0, 1]], [[1, 1, 0, 0], [0, 0, 0, 0]], 0.5, 1e-5],
    ],
)
def test_soft_jaccard_score_2(y_true, y_pred, expected, eps):
    y_true = torch.tensor(y_true, dtype=torch.float32)
    y_pred = torch.tensor(y_pred, dtype=torch.float32)
    actual = F.soft_jaccard_score(y_pred, y_true, dims=[1], eps=eps)
    actual = actual.mean()
    assert float(actual) == pytest.approx(expected, eps)


@pytest.mark.parametrize(
    ["y_true", "y_pred", "expected", "eps"],
    [
        [[1, 1, 1, 1], [1, 1, 1, 1], 1.0, 1e-5],
        [[0, 1, 1, 0], [0, 1, 1, 0], 1.0, 1e-5],
        [[1, 1, 1, 1], [1, 1, 0, 0], 2.0 / 3.0, 1e-5],
    ],
)
def test_soft_dice_score(y_true, y_pred, expected, eps):
    y_true = torch.tensor(y_true, dtype=torch.float32)
    y_pred = torch.tensor(y_pred, dtype=torch.float32)
    actual = F.soft_dice_score(y_pred, y_true, eps=eps)
    assert float(actual) == pytest.approx(expected, eps)


@torch.no_grad()
def test_dice_loss_binary():
    eps = 1e-5
    criterion = L.DiceLoss(mode="binary", from_logits=False)

    # Ideal case
    y_pred = torch.tensor([1.0, 1.0, 1.0]).view(1, 1, 1, -1)
    y_true = torch.tensor(([1, 1, 1])).view(1, 1, 1, -1)
    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(0.0, abs=eps)

    y_pred = torch.tensor([1.0, 0.0, 1.0]).view(1, 1, 1, -1)
    y_true = torch.tensor(([1, 0, 1])).view(1, 1, 1, -1)
    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(0.0, abs=eps)

    y_pred = torch.tensor([0.0, 0.0, 0.0]).view(1, 1, 1, -1)
    y_true = torch.tensor(([0, 0, 0])).view(1, 1, 1, -1)
    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(0.0, abs=eps)

    # Worst case
    y_pred = torch.tensor([1.0, 1.0, 1.0]).view(1, 1, -1)
    y_true = torch.tensor([0, 0, 0]).view(1, 1, 1, -1)
    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(0.0, abs=eps)

    y_pred = torch.tensor([1.0, 0.0, 1.0]).view(1, 1, -1)
    y_true = torch.tensor([0, 1, 0]).view(1, 1, 1, -1)
    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(1.0, abs=eps)

    y_pred = torch.tensor([0.0, 0.0, 0.0]).view(1, 1, -1)
    y_true = torch.tensor([1, 1, 1]).view(1, 1, 1, -1)
    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(1.0, abs=eps)


@torch.no_grad()
def test_binary_jaccard_loss():
    eps = 1e-5
    criterion = L.JaccardLoss(mode="binary", from_logits=False)

    # Ideal case
    y_pred = torch.tensor([1.0]).view(1, 1, 1, 1)
    y_true = torch.tensor(([1])).view(1, 1, 1, 1)
    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(0.0, abs=eps)

    y_pred = torch.tensor([1.0, 0.0, 1.0]).view(1, 1, 1, -1)
    y_true = torch.tensor(([1, 0, 1])).view(1, 1, 1, -1)
    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(0.0, abs=eps)

    y_pred = torch.tensor([0.0, 0.0, 0.0]).view(1, 1, 1, -1)
    y_true = torch.tensor(([0, 0, 0])).view(1, 1, 1, -1)
    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(0.0, abs=eps)

    # Worst case
    y_pred = torch.tensor([1.0, 1.0, 1.0]).view(1, 1, -1)
    y_true = torch.tensor([0, 0, 0]).view(1, 1, 1, -1)
    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(0.0, abs=eps)

    y_pred = torch.tensor([1.0, 0.0, 1.0]).view(1, 1, -1)
    y_true = torch.tensor([0, 1, 0]).view(1, 1, 1, -1)
    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(1.0, eps)

    y_pred = torch.tensor([0.0, 0.0, 0.0]).view(1, 1, -1)
    y_true = torch.tensor([1, 1, 1]).view(1, 1, 1, -1)
    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(1.0, eps)


@torch.no_grad()
def test_multiclass_jaccard_loss():
    eps = 1e-5
    criterion = L.JaccardLoss(mode="multiclass", from_logits=False)

    # Ideal case
    y_pred = torch.tensor([[[1.0, 1.0, 0.0, 0.0], [0.0, 0.0, 1.0, 1.0]]])
    y_true = torch.tensor([[0, 0, 1, 1]])

    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(0.0, abs=eps)

    # Worst case
    y_pred = torch.tensor([[[1.0, 1.0, 0.0, 0.0], [0.0, 0.0, 1.0, 1.0]]])
    y_true = torch.tensor([[1, 1, 0, 0]])

    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(1.0, abs=eps)

    # 1 - 1/3 case
    y_pred = torch.tensor([[[1.0, 0.0, 1.0, 0.0], [0.0, 1.0, 0.0, 1.0]]])
    y_true = torch.tensor([[1, 1, 0, 0]])

    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(1.0 - 1.0 / 3.0, abs=eps)


@torch.no_grad()
def test_multilabel_jaccard_loss():
    eps = 1e-5
    criterion = L.JaccardLoss(mode="multilabel", from_logits=False)

    # Ideal case
    y_pred = torch.tensor([[[1.0, 1.0, 0.0, 0.0], [0.0, 0.0, 1.0, 1.0]]])
    y_true = torch.tensor([[[1.0, 1.0, 0.0, 0.0], [0.0, 0.0, 1.0, 1.0]]])

    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(0.0, abs=eps)

    # Worst case
    y_pred = torch.tensor([[[1.0, 1.0, 0.0, 0.0], [0.0, 0.0, 1.0, 1.0]]])
    y_true = 1 - y_pred
    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(1.0, abs=eps)

    # 1 - 1/3 case
    y_pred = torch.tensor([[[0.0, 1.0, 1.0, 0.0], [0.0, 1.0, 1.0, 0.0]]])
    y_true = torch.tensor([[[1.0, 1.0, 0.0, 0.0], [1.0, 1.0, 0.0, 0.0]]])

    loss = criterion(y_pred, y_true)
    assert float(loss) == pytest.approx(1.0 - 1.0 / 3.0, abs=eps)


@torch.no_grad()
def test_soft_ce_loss():
    soft_ce_criterion = L.SoftCrossEntropyLoss(smooth_factor=0.0)
    ce_criterion = nn.CrossEntropyLoss()

    y_pred = torch.tensor([[+1, -1, -1, -1], [-1, +1, -1, -1], [-1, -1, +1, -1], [-1, -1, -1, +1]]).float()
    y_true = torch.tensor([0, 1, 2, 3]).long()

    actual = soft_ce_criterion(y_pred, y_true).item()
    expected = ce_criterion(y_pred, y_true).item()
    np.testing.assert_almost_equal(actual, expected)


@torch.no_grad()
def test_soft_bce_loss():
    criterion = L.SoftBCEWithLogitsLoss(smooth_factor=0.1, ignore_index=-100)

    # Ideal case
    y_pred = torch.tensor([-9, 9, 1, 9, -9]).float()
    y_true = torch.tensor([0, 1, -100, 1, 0]).long()

    loss = criterion(y_pred, y_true)
    print(loss)


@pytest.mark.parametrize(
    "criterion",
    [
        L.BiTemperedLogisticLoss(t1=1, t2=0.8),
        L.FocalCosineLoss(),
        L.BinaryFocalLoss(),
        L.CrossEntropyFocalLoss(),
        L.SoftF1Loss(),
        L.SoftCrossEntropyLoss(),
    ],
)
def test_classification_losses(criterion):
    # Ideal case
    y_pred = torch.tensor([[+9, -9, -9, -9], [-9, +9, -9, -9], [-9, -9, +9, -9], [-9, -9, -9, +9]]).float()
    y_true = torch.tensor([0, 1, 2, 3]).long()

    loss = criterion(y_pred, y_true)
    print(loss)


def test_binary_bi_tempered_loss():
    loss = L.BinaryBiTemperedLogisticLoss(t1=0.9, t2=3.0, ignore_index=-100)

    y_pred = torch.randn((4, 1, 512, 512))
    y_true = (y_pred > 0).type_as(y_pred)
    y_true[:, :, ::10, ::20] = -100
    loss_value = loss(y_pred, y_true)
    assert len(loss_value.size()) == 0


def test_bbce():
    x = torch.tensor([0, 0, 0, 0, 0]).float()
    y = torch.tensor([0, 1, 1, 1, 1]).float()
    loss = L.balanced_binary_cross_entropy_with_logits(x, y, gamma=1, reduction="none")
    print(loss)



import torch
from torch import nn, Tensor


class QualityFocalLoss(nn.Module):
    __constants__ = ["beta", "reduction"]

    def __init__(self, beta: float = 2, reduction="mean"):
        """
        Quality Focal Loss from https://arxiv.org/abs/2006.04388

        :param beta: Power factor for focal term
        :param reduction: Possible values are: mean, sum, normalized
            mean - mean loss value is returned
            sum - sum of loss values is returned
            normalized - mean loss value is divided by sum of focal terms.
            This mimics normalized focal loss approach.
        """
        super().__init__()
        self.beta = beta
        self.reduction = reduction

    @torch.cuda.amp.autocast(False)
    def forward(self, predictions: Tensor, targets: Tensor) -> Tensor:
        """
        Compute quality focal loss

        :param predictions: Prediction logits
        :param targets: Targets of the same shape as predictions
        :return: Loss value
        """
        predictions = predictions.float()
        targets = targets.float()

        bce = torch.nn.functional.binary_cross_entropy_with_logits(predictions, targets, reduction="none")
        focal_term = torch.nn.functional.l1_loss(predictions.sigmoid(), targets, reduction="none").pow_(self.beta)
        loss = focal_term * bce

        if self.reduction == "mean":
            return loss.mean()
        if self.reduction == "sum":
            return loss.sum()
        if self.reduction == "normalized":
            return loss.sum() / focal_term.sum()

        return loss



import numpy as np

from pytorch_toolbelt.utils.bucket_assignment import (
    naive_bucket_assignment,
    compute_bucket_imbalance_score,
    random_bucket_assignment,
    filler_bucket_assignment,
)


def test_approximate_bucket_assignment():
    cost = np.concatenate(
        [
            np.random.randint(0, 100, size=100),
            np.random.randint(100, 1000, size=50),
            np.random.randint(1000, 5000, size=5),
            np.random.randint(10000, 20000, size=1),
            np.random.randint(20000, 100000, size=3),
        ]
    )
    cost = np.random.permutation(cost)

    assignment = naive_bucket_assignment(cost, 4)

    print("naive_bucket_assignment  ", compute_bucket_imbalance_score(cost, assignment))
    print(np.bincount(assignment))

    assignment = random_bucket_assignment(cost, 4, 1000)

    print("random_bucket_assignment ", compute_bucket_imbalance_score(cost, assignment))
    print(np.bincount(assignment))

    assignment = filler_bucket_assignment(cost, 4)

    print("filler_bucket_assignment ", compute_bucket_imbalance_score(cost, assignment))
    print(np.bincount(assignment))



import math
import warnings


import torch
from torch import nn

from ..common import EncoderModule, _take
from ...activations import ACT_SILU, get_activation_block

__all__ = [
    "TimmB0Encoder",
    "TimmB1Encoder",
    "TimmB2Encoder",
    "TimmB3Encoder",
    "TimmB4Encoder",
    "TimmB5Encoder",
    "TimmB6Encoder",
    "TimmB7Encoder",
    "TimmMixNetXLEncoder",
    "B0Encoder",
    "B1Encoder",
    "B2Encoder",
    "B3Encoder",
    "B4Encoder",
    "B5Encoder",
    "B6Encoder",
    "B7Encoder",
    "MixNetXLEncoder",
]


def make_n_channel_input_conv2d_same(conv: nn.Conv2d, in_channels: int, mode="auto", **kwargs):
    assert isinstance(conv, nn.Conv2d)
    if conv.in_channels == in_channels:
        warnings.warn("make_n_channel_input call is spurious")
        return conv

    from timm.models.layers import Conv2dSame

    new_conv = Conv2dSame(
        in_channels,
        out_channels=conv.out_channels,
        kernel_size=kwargs.get("kernel_size", conv.kernel_size),
        stride=kwargs.get("stride", conv.stride),
        padding=kwargs.get("padding", conv.padding),
        dilation=kwargs.get("dilation", conv.dilation),
        groups=kwargs.get("groups", conv.groups),
        bias=kwargs.get("bias", conv.bias is not None),
    )

    w = conv.weight
    if in_channels > conv.in_channels:
        n = math.ceil(in_channels / float(conv.in_channels))
        w = torch.cat([w] * n, dim=1)
        w = w[:, :in_channels, ...]
        new_conv.weight = nn.Parameter(w, requires_grad=True)
    else:
        w = w[:, 0:in_channels, ...]
        new_conv.weight = nn.Parameter(w, requires_grad=True)

    return new_conv


class TimmBaseEfficientNetEncoder(EncoderModule):
    def __init__(self, encoder, features, layers=[1, 2, 3, 4], first_conv_stride_one: bool = False):
        strides = [2, 4, 8, 16, 32]

        if first_conv_stride_one:
            strides = [1, 2, 4, 8, 16]
            encoder.conv_stem.stride = (1, 1)

        # if no_stride_s16:
        #     encoder.blocks[3][0].conv_dw.stride = (1, 1)
        #     encoder.blocks[3][0].conv_dw.dilation = (2, 2)
        #     strides[3] = strides[2]
        #     strides[4] = strides[3] * 2
        #
        # if no_stride_s32:
        #     encoder.blocks[5][0].conv_dw.stride = (1, 1)
        #     encoder.blocks[5][0].conv_dw.dilation = (2, 2)
        #     strides[4] = strides[3]

        super().__init__(features, strides, layers)
        self.encoder = encoder

    def forward(self, x):
        features = self.encoder(x)
        return _take(features, self._layers)

    @torch.jit.unused
    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.conv_stem = make_n_channel_input_conv2d_same(
            self.encoder.conv_stem, input_channels, mode, **kwargs
        )
        return self


class TimmB0Encoder(TimmBaseEfficientNetEncoder):
    def __init__(
        self,
        pretrained=True,
        layers=[1, 2, 3, 4],
        activation: str = ACT_SILU,
        first_conv_stride_one: bool = False,
        use_tf=True,
    ):
        from timm.models.efficientnet import tf_efficientnet_b0_ns, efficientnet_b0

        model_cls = tf_efficientnet_b0_ns if use_tf else efficientnet_b0

        act_layer = get_activation_block(activation)
        encoder = model_cls(pretrained=pretrained, features_only=True, act_layer=act_layer, drop_path_rate=0.05)

        super().__init__(
            encoder, features=[16, 24, 40, 112, 320], layers=layers, first_conv_stride_one=first_conv_stride_one
        )


class TimmB1Encoder(TimmBaseEfficientNetEncoder):
    def __init__(
        self,
        pretrained: bool = True,
        layers=[1, 2, 3, 4],
        activation: str = ACT_SILU,
        first_conv_stride_one: bool = False,
        drop_path_rate: float = 0.05,
        **kwargs,
    ):
        from timm.models.efficientnet import tf_efficientnet_b1_ns

        act_layer = get_activation_block(activation)
        encoder = tf_efficientnet_b1_ns(
            pretrained=pretrained,
            features_only=True,
            act_layer=act_layer,
            drop_path_rate=drop_path_rate,
            **kwargs,
        )
        super().__init__(encoder, [16, 24, 40, 112, 320], layers=layers, first_conv_stride_one=first_conv_stride_one)


class TimmB2Encoder(TimmBaseEfficientNetEncoder):
    def __init__(
        self,
        pretrained=True,
        layers=[1, 2, 3, 4],
        activation: str = ACT_SILU,
        first_conv_stride_one: bool = False,
        drop_path_rate: float = 0.1,
    ):
        from timm.models.efficientnet import tf_efficientnet_b2_ns

        act_layer = get_activation_block(activation)
        encoder = tf_efficientnet_b2_ns(
            pretrained=pretrained, features_only=True, act_layer=act_layer, drop_path_rate=drop_path_rate
        )
        super().__init__(encoder, [16, 24, 48, 120, 352], layers=layers, first_conv_stride_one=first_conv_stride_one)


class TimmB3Encoder(TimmBaseEfficientNetEncoder):
    def __init__(
        self,
        pretrained=True,
        layers=[1, 2, 3, 4],
        activation: str = ACT_SILU,
        first_conv_stride_one: bool = False,
        drop_path_rate=0.1,
    ):
        from timm.models.efficientnet import tf_efficientnet_b3_ns

        act_layer = get_activation_block(activation)
        encoder = tf_efficientnet_b3_ns(
            pretrained=pretrained, features_only=True, act_layer=act_layer, drop_path_rate=drop_path_rate
        )
        super().__init__(encoder, [24, 32, 48, 136, 384], layers=layers, first_conv_stride_one=first_conv_stride_one)


class TimmB4Encoder(TimmBaseEfficientNetEncoder):
    def __init__(
        self,
        pretrained=True,
        layers=[1, 2, 3, 4],
        activation: str = ACT_SILU,
        first_conv_stride_one: bool = False,
        drop_path_rate=0.2,
    ):
        from timm.models.efficientnet import tf_efficientnet_b4_ns

        act_layer = get_activation_block(activation)
        encoder = tf_efficientnet_b4_ns(
            pretrained=pretrained, features_only=True, act_layer=act_layer, drop_path_rate=drop_path_rate
        )

        super().__init__(encoder, [24, 32, 56, 160, 448], layers=layers, first_conv_stride_one=first_conv_stride_one)


class TimmB5Encoder(TimmBaseEfficientNetEncoder):
    def __init__(
        self,
        pretrained=True,
        layers=[1, 2, 3, 4],
        activation: str = ACT_SILU,
        first_conv_stride_one: bool = False,
        drop_path_rate=0.2,
    ):
        from timm.models.efficientnet import tf_efficientnet_b5_ns

        act_layer = get_activation_block(activation)
        encoder = tf_efficientnet_b5_ns(
            pretrained=pretrained, features_only=True, act_layer=act_layer, drop_path_rate=drop_path_rate
        )
        super().__init__(encoder, [24, 40, 64, 176, 512], layers=layers, first_conv_stride_one=first_conv_stride_one)


class TimmB6Encoder(TimmBaseEfficientNetEncoder):
    def __init__(
        self,
        pretrained=True,
        layers=[1, 2, 3, 4],
        activation: str = ACT_SILU,
        first_conv_stride_one: bool = False,
        drop_path_rate=0.2,
    ):
        from timm.models.efficientnet import tf_efficientnet_b6_ns

        act_layer = get_activation_block(activation)
        encoder = tf_efficientnet_b6_ns(
            pretrained=pretrained, features_only=True, act_layer=act_layer, drop_path_rate=drop_path_rate
        )
        super().__init__(encoder, [32, 40, 72, 200, 576], layers=layers, first_conv_stride_one=first_conv_stride_one)


class TimmB7Encoder(TimmBaseEfficientNetEncoder):
    def __init__(
        self,
        pretrained=True,
        layers=[1, 2, 3, 4],
        activation: str = ACT_SILU,
        first_conv_stride_one: bool = False,
        drop_path_rate=0.2,
    ):
        from timm.models.efficientnet import tf_efficientnet_b7_ns

        act_layer = get_activation_block(activation)
        encoder = tf_efficientnet_b7_ns(
            pretrained=pretrained, features_only=True, act_layer=act_layer, drop_path_rate=drop_path_rate
        )
        super().__init__(encoder, [32, 48, 80, 224, 640], layers=layers, first_conv_stride_one=first_conv_stride_one)


class TimmMixNetXLEncoder(TimmBaseEfficientNetEncoder):
    def __init__(
        self,
        pretrained=True,
        layers=[1, 2, 3, 4],
        activation: str = ACT_SILU,
        first_conv_stride_one: bool = False,
        drop_path_rate=0.2,
    ):
        from timm.models.efficientnet import mixnet_xl

        act_layer = get_activation_block(activation)
        encoder = mixnet_xl(
            pretrained=pretrained, features_only=True, act_layer=act_layer, drop_path_rate=drop_path_rate
        )
        super().__init__(encoder, [40, 48, 64, 192, 320], layers=layers, first_conv_stride_one=first_conv_stride_one)


# Aliases to keep backward compatibility
B0Encoder = TimmB0Encoder
B1Encoder = TimmB1Encoder
B2Encoder = TimmB2Encoder
B3Encoder = TimmB3Encoder
B4Encoder = TimmB4Encoder
B5Encoder = TimmB5Encoder
B6Encoder = TimmB6Encoder
B7Encoder = TimmB7Encoder
MixNetXLEncoder = TimmMixNetXLEncoder
